"""
==============================================================
Crossing invariant fiber response function with FORECAST model
==============================================================

We show how to obtain a voxel specific response function in the form of
axially symmetric tensor and the fODF using the FORECAST model from
:footcite:p:`Anderson2005`, :footcite:p:`Kaden2016a` and
:footcite:p:`Zucchelli2017`.

First import the necessary modules:
"""

import os.path as op

import matplotlib.pyplot as plt
import nibabel as nib
import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import fetch_hbn, get_sphere
from dipy.reconst.forecast import ForecastModel
from dipy.viz import actor, window

###############################################################################
# Download and read the data for this tutorial. Our implementation of FORECAST
# requires multi-shell `data.fetch_hbn()` provides data that was acquired using
# b-values of 1000 and 2000 as part of the Healthy Brain Network study
# :footcite:p:`Alexander2017` and was preprocessed and quality controlled in the
# HBN-POD2 dataset :footcite:p:`RichieHalford2022`.

data_path = fetch_hbn(["NDARAA948VFH"])[1]
dwi_path = op.join(
    data_path, "derivatives", "qsiprep", "sub-NDARAA948VFH", "ses-HBNsiteRU", "dwi"
)

img = nib.load(
    op.join(
        dwi_path,
        "sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi.nii.gz",
    )
)

gtab = gradient_table(
    op.join(
        dwi_path,
        "sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi.bval",
    ),
    bvecs=op.join(
        dwi_path,
        "sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi.bvec",
    ),
)

data = np.asarray(img.dataobj)

mask_img = nib.load(
    op.join(
        dwi_path,
        "sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-brain_mask.nii.gz",
    )
)

brain_mask = mask_img.get_fdata()

###############################################################################
# Let us consider only a single slice for the FORECAST fitting

data_small = data[:, :, 50:51]
mask_small = brain_mask[:, :, 50:51]

###############################################################################
# Instantiate the FORECAST Model.
#
# "sh_order_max" is the spherical harmonics order (l) used for the fODF.
#
# dec_alg is the spherical deconvolution algorithm used for the FORECAST basis
# fitting, in this case we used the Constrained Spherical Deconvolution (CSD)
# algorithm.

fm = ForecastModel(gtab, sh_order_max=6, dec_alg="CSD")

###############################################################################
# Fit the FORECAST to the data

f_fit = fm.fit(data_small, mask=mask_small)

###############################################################################
# Calculate the crossing invariant tensor indices :footcite:p:`Kaden2016a`: the
# parallel diffusivity, the perpendicular diffusivity, the fractional anisotropy
# and the mean diffusivity.

d_par = f_fit.dpar
d_perp = f_fit.dperp
fa = f_fit.fractional_anisotropy()
md = f_fit.mean_diffusivity()

###############################################################################
# Show the indices and save them in FORECAST_indices.png.

fig = plt.figure(figsize=(6, 6))
ax1 = fig.add_subplot(2, 2, 1, title="parallel diffusivity")
ax1.set_axis_off()
ind = ax1.imshow(
    d_par[:, :, 0].T, interpolation="nearest", origin="lower", cmap=plt.cm.gray
)
plt.colorbar(ind, shrink=0.6)
ax2 = fig.add_subplot(2, 2, 2, title="perpendicular diffusivity")
ax2.set_axis_off()
ind = ax2.imshow(
    d_perp[:, :, 0].T, interpolation="nearest", origin="lower", cmap=plt.cm.gray
)
plt.colorbar(ind, shrink=0.6)
ax3 = fig.add_subplot(2, 2, 3, title="fractional anisotropy")
ax3.set_axis_off()
ind = ax3.imshow(
    fa[:, :, 0].T, interpolation="nearest", origin="lower", cmap=plt.cm.gray
)
plt.colorbar(ind, shrink=0.6)
ax4 = fig.add_subplot(2, 2, 4, title="mean diffusivity")
ax4.set_axis_off()
ind = ax4.imshow(
    md[:, :, 0].T, interpolation="nearest", origin="lower", cmap=plt.cm.gray
)
plt.colorbar(ind, shrink=0.6)
plt.savefig("FORECAST_indices.png", dpi=300, bbox_inches="tight")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# FORECAST scalar indices.
#
#
# Load an ODF reconstruction sphere

sphere = get_sphere(name="repulsion724")

###############################################################################
# Compute the fODFs.

odf = f_fit.odf(sphere)
print(f"fODF.shape {odf.shape}")

###############################################################################
# Display a part of the fODFs

odf_actor = actor.odf_slicer(
    odf[30:60, 30:60, :], sphere=sphere, colormap="plasma", scale=0.6
)
scene = window.Scene()
scene.add(odf_actor)
window.record(scene=scene, out_path="fODFs.png", size=(600, 600), magnification=4)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Fiber Orientation Distribution Functions, in a small ROI of the brain.
#
#
# References
# ----------
#
# .. footbibliography::
#



#!python
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True

import numpy as np
cimport numpy as cnp

cdef extern from "dpy_math.h" nogil:
    double cos(double)
    double sin(double)
    double log(double)

cdef class Transform:
    r""" Base class (contract) for all transforms for affine image registration
    Each transform must define the following (fast, nogil) methods:

    1. _jacobian(theta, x, J): receives a parameter vector theta, a point in
       x, and a matrix J with shape (dim, len(theta)). It must writes in J, the
       Jacobian of the transform with parameters theta evaluated at x.

    2. _get_identity_parameters(theta): receives a vector theta whose length is
       the number of parameters of the transform and sets in theta the values
       that define the identity transform.

    3. _param_to_matrix(theta, T): receives a parameter vector theta, and a
       matrix T of shape (dim + 1, dim + 1) and writes in T the matrix
       representation of the transform with parameters theta

    This base class defines the (slow, convenient) python wrappers for each
    of the above functions, which also do parameter checking and raise
    a ValueError in case the provided parameters are invalid.
    """
    def __cinit__(self):
        r""" Default constructor
        Sets transform dimension and number of parameter to invalid values (-1)
        """
        self.dim = -1
        self.number_of_parameters = -1

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        return -1

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        return

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] T) noexcept nogil:
        return

    def jacobian(self, double[:] theta, double[:] x):
        r""" Jacobian function of this transform

        Parameters
        ----------
        theta : array, shape (n,)
            vector containing the n parameters of this transform
        x : array, shape (dim,)
            vector containing the point where the Jacobian must be evaluated

        Returns
        -------
        J : array, shape (dim, n)
            Jacobian matrix of the transform with parameters theta at point x
        """
        n = theta.shape[0]
        if n != self.number_of_parameters:
            raise ValueError("Invalid number of parameters: %d"%(n,))
        m = x.shape[0]
        if m < self.dim:
            raise ValueError("Invalid point dimension: %d"%(m,))
        J = np.zeros((self.dim, n))
        ret = self._jacobian(theta, x, J)
        return np.asarray(J)

    def get_identity_parameters(self):
        r""" Parameter values corresponding to the identity transform

        Returns
        -------
        theta : array, shape (n,)
            the n parameter values corresponding to the identity transform
        """
        if self.number_of_parameters < 0:
            raise ValueError("Invalid transform.")
        theta = np.zeros(self.number_of_parameters)
        self._get_identity_parameters(theta)
        return np.asarray(theta)

    def param_to_matrix(self, double[:] theta):
        r""" Matrix representation of this transform with the given parameters

        Parameters
        ----------
        theta : array, shape (n,)
            the parameter values of the transform

        Returns
        -------
        T : array, shape (dim + 1, dim + 1)
            the matrix representation of this transform with parameters theta
        """
        n = len(theta)
        if n != self.number_of_parameters:
            raise ValueError("Invalid number of parameters: %d"%(n,))
        T = np.eye(self.dim + 1)
        self._param_to_matrix(theta, T)
        return np.asarray(T)

    def get_number_of_parameters(self):
        return self.number_of_parameters

    def get_dim(self):
        return self.dim


cdef class TranslationTransform2D(Transform):
    def __init__(self):
        r""" Translation transform in 2D
        """
        self.dim = 2
        self.number_of_parameters = 2

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        r""" Jacobian matrix of the 2D translation transform
        The transformation is given by:

        T(x) = (T1(x), T2(x)) = (x0 + t0, x1 + t1)

        The derivative w.r.t. t1 and t2 is given by

        T'(x) = [[1, 0], # derivatives of [T1, T2] w.r.t. t0
                 [0, 1]] # derivatives of [T1, T2] w.r.t. t1

        Parameters
        ----------
        theta : array, shape (2,)
            the parameters of the 2D translation transform (the Jacobian does
            not depend on the parameters, but we receive the buffer so all
            Jacobian functions receive the same parameters)
        x : array, shape (2,)
            the point at which to compute the Jacobian (the Jacobian does not
            depend on x, but we receive the buffer so all Jacobian functions
            receive the same parameters)
        J : array, shape (2, 2)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 1, indicating that the Jacobian is constant
            (independent of x)
        """
        J[0, 0], J[0, 1] = 1.0, 0.0
        J[1, 0], J[1, 1] = 0.0, 1.0
        # This Jacobian does not depend on x (it's constant): return 1
        return 1

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        r""" Parameter values corresponding to the identity
        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (2,)
            buffer to write the parameters of the 2D translation transform
        """
        theta[:2] = 0

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        r""" Matrix associated with the 2D translation transform

        Parameters
        ----------
        theta : array, shape (2,)
            the parameters of the 2D translation transform
        R : array, shape (3, 3)
            the buffer in which to write the translation matrix
        """
        R[0, 0], R[0, 1], R[0, 2] = 1, 0, theta[0]
        R[1, 0], R[1, 1], R[1, 2] = 0, 1, theta[1]
        R[2, 0], R[2, 1], R[2, 2] = 0, 0, 1


cdef class TranslationTransform3D(Transform):
    def __init__(self):
        r""" Translation transform in 3D
        """
        self.dim = 3
        self.number_of_parameters = 3

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        r""" Jacobian matrix of the 3D translation transform
        The transformation is given by:

        T(x) = (T1(x), T2(x), T3(x)) = (x0 + t0, x1 + t1, x2 + t2)

        The derivative w.r.t. t1, t2 and t3 is given by

        T'(x) = [[1, 0, 0], # derivatives of [T1, T2, T3] w.r.t. t0
                 [0, 1, 0], # derivatives of [T1, T2, T3] w.r.t. t1
                 [0, 0, 1]] # derivatives of [T1, T2, T3] w.r.t. t2
        Parameters
        ----------
        theta : array, shape (3,)
            the parameters of the 3D translation transform (the Jacobian does
            not depend on the parameters, but we receive the buffer so all
            Jacobian functions receive the same parameters)
        x : array, shape (3,)
            the point at which to compute the Jacobian (the Jacobian does not
            depend on x, but we receive the buffer so all Jacobian functions
            receive the same parameters)
        J : array, shape (3, 3)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 1, indicating that the Jacobian is constant
            (independent of x)
        """
        J[0, 0], J[0, 1], J[0, 2] = 1.0, 0.0, 0.0
        J[1, 0], J[1, 1], J[1, 2] = 0.0, 1.0, 0.0
        J[2, 0], J[2, 1], J[2, 2] = 0.0, 0.0, 1.0
        # This Jacobian does not depend on x (it's constant): return 1
        return 1

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        r""" Parameter values corresponding to the identity
        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (3,)
            buffer to write the parameters of the 3D translation transform
        """
        theta[:3] = 0

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        r""" Matrix associated with the 3D translation transform

        Parameters
        ----------
        theta : array, shape (3,)
            the parameters of the 3D translation transform
        R : array, shape (4, 4)
            the buffer in which to write the translation matrix
        """
        R[0, 0], R[0, 1], R[0, 2], R[0, 3] = 1, 0, 0, theta[0]
        R[1, 0], R[1, 1], R[1, 2], R[1, 3] = 0, 1, 0, theta[1]
        R[2, 0], R[2, 1], R[2, 2], R[2, 3] = 0, 0, 1, theta[2]
        R[3, 0], R[3, 1], R[3, 2], R[3, 3] = 0, 0, 0, 1


cdef class RotationTransform2D(Transform):
    def __init__(self):
        r""" Rotation transform in 2D
        """
        self.dim = 2
        self.number_of_parameters = 1

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        r""" Jacobian matrix of a 2D rotation with parameter theta, at x

        The transformation is given by:

        T(x,y) = (T1(x,y), T2(x,y)) = (x cost - y sint, x sint + y cost)

        The derivatives w.r.t. the rotation angle, t, are:

        T'(x,y) = [-x sint - y cost, # derivative of T1 w.r.t. t
                    x cost - y sint] # derivative of T2 w.r.t. t

        Parameters
        ----------
        theta : array, shape (1,)
            the rotation angle
        x : array, shape (2,)
            the point at which to compute the Jacobian
        J : array, shape (2, 1)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 0, indicating that the Jacobian is not
            constant (it depends on the value of x)
        """
        cdef:
            double st = sin(theta[0])
            double ct = cos(theta[0])
            double px = x[0], py = x[1]

        J[0, 0] = -px * st - py * ct
        J[1, 0] = px * ct - py * st
        # This Jacobian depends on x (it's not constant): return 0
        return 0

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        r""" Parameter values corresponding to the identity
        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (1,)
            buffer to write the parameters of the 2D rotation transform
        """
        theta[0] = 0

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        r""" Matrix associated with the 2D rotation transform

        Parameters
        ----------
        theta : array, shape (1,)
            the rotation angle
        R : array, shape (3,3)
            the buffer in which to write the matrix
        """
        cdef:
            double ct = cos(theta[0])
            double st = sin(theta[0])
        R[0, 0], R[0, 1], R[0, 2] = ct, -st, 0
        R[1, 0], R[1, 1], R[1, 2] = st, ct, 0
        R[2, 0], R[2, 1], R[2, 2] = 0, 0, 1


cdef class RotationTransform3D(Transform):
    def __init__(self):
        r""" Rotation transform in 3D
        """
        self.dim = 3
        self.number_of_parameters = 3

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        r""" Jacobian matrix of a 3D rotation with parameters theta, at x

        Parameters
        ----------
        theta : array, shape (3,)
            the rotation angles about the canonical axes
        x : array, shape (3,)
            the point at which to compute the Jacobian
        J : array, shape (3, 3)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 0, indicating that the Jacobian is not
            constant (it depends on the value of x)
        """
        cdef:
            double sa = sin(theta[0])
            double ca = cos(theta[0])
            double sb = sin(theta[1])
            double cb = cos(theta[1])
            double sc = sin(theta[2])
            double cc = cos(theta[2])
            double px = x[0], py = x[1], z = x[2]

        J[0, 0] = (-sc * ca * sb) * px + (sc * sa) * py + (sc * ca * cb) * z
        J[1, 0] = (cc * ca * sb) * px + (-cc * sa) * py + (-cc * ca * cb) * z
        J[2, 0] = (sa * sb) * px + ca * py + (-sa * cb) * z

        J[0, 1] = (-cc * sb - sc * sa * cb) * px + (cc * cb - sc * sa * sb) * z
        J[1, 1] = (-sc * sb + cc * sa * cb) * px + (sc * cb + cc * sa * sb) * z
        J[2, 1] = (-ca * cb) * px + (-ca * sb) * z

        J[0, 2] = (-sc * cb - cc * sa * sb) * px + (-cc * ca) * py + \
                  (-sc * sb + cc * sa * cb) * z
        J[1, 2] = (cc * cb - sc * sa * sb) * px + (-sc * ca) * py + \
                  (cc * sb + sc * sa * cb) * z
        J[2, 2] = 0
        # This Jacobian depends on x (it's not constant): return 0
        return 0

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        r""" Parameter values corresponding to the identity
        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (3,)
            buffer to write the parameters of the 3D rotation transform
        """
        theta[:3] = 0

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        r""" Matrix associated with the 3D rotation transform

        The matrix is the product of rotation matrices of angles theta[0],
        theta[1], theta[2] around axes x, y, z applied in the following
        order: y, x, z. This order was chosen for consistency with ANTS.

        Parameters
        ----------
        theta : array, shape (3,)
            the rotation angles about each axis:
            theta[0] : rotation angle around x axis
            theta[1] : rotation angle around y axis
            theta[2] : rotation angle around z axis
        R : array, shape (4, 4)
            buffer in which to write the rotation matrix
        """
        cdef:
            double sa = sin(theta[0])
            double ca = cos(theta[0])
            double sb = sin(theta[1])
            double cb = cos(theta[1])
            double sc = sin(theta[2])
            double cc = cos(theta[2])

        R[0,0], R[0,1], R[0,2] = cc*cb-sc*sa*sb, -sc*ca, cc*sb+sc*sa*cb
        R[1,0], R[1,1], R[1,2] = sc*cb+cc*sa*sb, cc*ca, sc*sb-cc*sa*cb
        R[2,0], R[2,1], R[2,2] = -ca*sb, sa, ca*cb
        R[3,0], R[3,1], R[3,2] = 0, 0, 0
        R[0, 3] = 0
        R[1, 3] = 0
        R[2, 3] = 0
        R[3, 3] = 1


cdef class RigidTransform2D(Transform):
    def __init__(self):
        r""" Rigid transform in 2D (rotation + translation)
        The parameter vector theta of length 3 is interpreted as follows:
        theta[0] : rotation angle
        theta[1] : translation along the x axis
        theta[2] : translation along the y axis
        """
        self.dim = 2
        self.number_of_parameters = 3

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        r""" Jacobian matrix of a 2D rigid transform (rotation + translation)

        The transformation is given by:

        T(x,y) = (T1(x,y), T2(x,y)) =
                 (x cost - y sint + dx, x sint + y cost + dy)

        The derivatives w.r.t. t, dx and dy are:

        T'(x,y) = [-x sint - y cost, 1, 0, # derivative of T1 w.r.t. t, dx, dy
                    x cost - y sint, 0, 1] # derivative of T2 w.r.t. t, dx, dy

        Parameters
        ----------
        theta : array, shape (3,)
            the parameters of the 2D rigid transform
            theta[0] : rotation angle (t)
            theta[1] : translation along the x axis (dx)
            theta[2] : translation along the y axis (dy)
        x : array, shape (2,)
            the point at which to compute the Jacobian
        J : array, shape (2, 3)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 0, indicating that the Jacobian is not
            constant (it depends on the value of x)
        """
        cdef:
            double st = sin(theta[0])
            double ct = cos(theta[0])
            double px = x[0], py = x[1]

        J[0, 0], J[0, 1], J[0, 2] = -px * st - py * ct, 1, 0
        J[1, 0], J[1, 1], J[1, 2] = px * ct - py * st, 0, 1
        # This Jacobian depends on x (it's not constant): return 0
        return 0

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        r""" Parameter values corresponding to the identity
        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (3,)
            buffer to write the parameters of the 2D rigid transform
            theta[0] : rotation angle
            theta[1] : translation along the x axis
            theta[2] : translation along the y axis
        """
        theta[:3] = 0

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        r""" Matrix associated with the 2D rigid transform

        Parameters
        ----------
        theta : array, shape (3,)
            the parameters of the 2D rigid transform
            theta[0] : rotation angle
            theta[1] : translation along the x axis
            theta[2] : translation along the y axis
        R : array, shape (3, 3)
            buffer in which to write the rigid matrix
        """
        cdef:
            double ct = cos(theta[0])
            double st = sin(theta[0])
        R[0, 0], R[0, 1], R[0, 2] = ct, -st, theta[1]
        R[1, 0], R[1, 1], R[1, 2] = st, ct, theta[2]
        R[2, 0], R[2, 1], R[2, 2] = 0, 0, 1


cdef class RigidTransform3D(Transform):
    def __init__(self):
        r""" Rigid transform in 3D (rotation + translation)
        The parameter vector theta of length 6 is interpreted as follows:
        theta[0] : rotation about the x axis
        theta[1] : rotation about the y axis
        theta[2] : rotation about the z axis
        theta[3] : translation along the x axis
        theta[4] : translation along the y axis
        theta[5] : translation along the z axis
        """
        self.dim = 3
        self.number_of_parameters = 6

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        r""" Jacobian matrix of a 3D rigid transform (rotation + translation)

        Parameters
        ----------
        theta : array, shape (6,)
            the parameters of the 3D rigid transform
            theta[0] : rotation about the x axis
            theta[1] : rotation about the y axis
            theta[2] : rotation about the z axis
            theta[3] : translation along the x axis
            theta[4] : translation along the y axis
            theta[5] : translation along the z axis
        x : array, shape (3,)
            the point at which to compute the Jacobian
        J : array, shape (3, 6)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 0, indicating that the Jacobian is not
            constant (it depends on the value of x)
        """
        cdef:
            double sa = sin(theta[0])
            double ca = cos(theta[0])
            double sb = sin(theta[1])
            double cb = cos(theta[1])
            double sc = sin(theta[2])
            double cc = cos(theta[2])
            double px = x[0], py = x[1], pz = x[2]

        J[0, 0] = (-sc * ca * sb) * px + (sc * sa) * py + (sc * ca * cb) * pz
        J[1, 0] = (cc * ca * sb) * px + (-cc * sa) * py + (-cc * ca * cb) * pz
        J[2, 0] = (sa * sb) * px + ca * py + (-sa * cb) * pz

        J[0, 1] = (-cc * sb - sc * sa * cb) * px + (cc * cb - sc * sa * sb) * pz
        J[1, 1] = (-sc * sb + cc * sa * cb) * px + (sc * cb + cc * sa * sb) * pz
        J[2, 1] = (-ca * cb) * px + (-ca * sb) * pz

        J[0, 2] = (-sc * cb - cc * sa * sb) * px + (-cc * ca) * py + \
                  (-sc * sb + cc * sa * cb) * pz
        J[1, 2] = (cc * cb - sc * sa * sb) * px + (-sc * ca) * py + \
                  (cc * sb + sc * sa * cb) * pz
        J[2, 2] = 0

        J[0, 3:6] = 0
        J[1, 3:6] = 0
        J[2, 3:6] = 0
        J[0, 3], J[1, 4], J[2, 5] = 1, 1, 1
        # This Jacobian depends on x (it's not constant): return 0
        return 0

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        r""" Parameter values corresponding to the identity
        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (6,)
            buffer to write the parameters of the 3D rigid transform
            theta[0] : rotation about the x axis
            theta[1] : rotation about the y axis
            theta[2] : rotation about the z axis
            theta[3] : translation along the x axis
            theta[4] : translation along the y axis
            theta[5] : translation along the z axis
        """
        theta[:6] = 0

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        r""" Matrix associated with the 3D rigid transform

        Parameters
        ----------
        theta : array, shape (6,)
            the parameters of the 3D rigid transform
            theta[0] : rotation about the x axis
            theta[1] : rotation about the y axis
            theta[2] : rotation about the z axis
            theta[3] : translation along the x axis
            theta[4] : translation along the y axis
            theta[5] : translation along the z axis
        R : array, shape (4, 4)
            buffer in which to write the rigid matrix
        """
        cdef:
            double sa = sin(theta[0])
            double ca = cos(theta[0])
            double sb = sin(theta[1])
            double cb = cos(theta[1])
            double sc = sin(theta[2])
            double cc = cos(theta[2])
            double dx = theta[3]
            double dy = theta[4]
            double dz = theta[5]

        R[0,0], R[0,1], R[0,2] = cc*cb-sc*sa*sb, -sc*ca, cc*sb+sc*sa*cb
        R[1,0], R[1,1], R[1,2] = sc*cb+cc*sa*sb, cc*ca, sc*sb-cc*sa*cb
        R[2,0], R[2,1], R[2,2] = -ca*sb, sa, ca*cb
        R[3,0], R[3,1], R[3,2] = 0, 0, 0
        R[0,3] = dx
        R[1,3] = dy
        R[2,3] = dz
        R[3,3] = 1


cdef class RigidIsoScalingTransform2D(Transform):
    def __init__(self):
        """ Rigid isoscaling transform in 2D.

        (rotation + translation + scaling)

        The parameter vector theta of length 4 is interpreted as follows:
        theta[0] : rotation angle
        theta[1] : translation along the x axis
        theta[2] : translation along the y axis
        theta[3] : isotropic scaling
        """
        self.dim = 2
        self.number_of_parameters = 4

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        r""" Jacobian matrix of a 2D rigid isoscaling transform.

        The transformation (rotation + translation + isoscaling) is given by:

        T(x,y) = (T1(x,y), T2(x,y)) =
            (sx * x * cost -  sx * y * sint + dx, sy * x * sint + sy * y * cost + dy)

        Parameters
        ----------
        theta : array, shape (4,)
            the parameters of the 2D rigid isoscaling transform
            theta[0] : rotation angle (t)
            theta[1] : translation along the x axis (dx)
            theta[2] : translation along the y axis (dy)
            theta[3] : isotropic scaling (s)
        x : array, shape (2,)
            the point at which to compute the Jacobian
        J : array, shape (2, 4)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 0, indicating that the Jacobian is not
            constant (it depends on the value of x)
        """
        cdef:
            double st = sin(theta[0])
            double ct = cos(theta[0])
            double px = x[0], py = x[1]
            double scale = theta[3]

        J[0, 0], J[0, 1], J[0, 2] = -px * scale * st - py * scale * ct, 1, 0
        J[1, 0], J[1, 1], J[1, 2] = px * scale * ct - py * scale * st, 0, 1

        J[0, 3] = px * ct - py * st
        J[1, 3] = px * st + py * ct
        # This Jacobian depends on x (it's not constant): return 0
        return 0

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        """ Parameter values corresponding to the identity
        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (4,)
            buffer to write the parameters of the 2D isoscaling rigid transform
            theta[0] : rotation angle
            theta[1] : translation along the x axis
            theta[2] : translation along the y axis
            theta[3] : isotropic scaling
        """
        theta[:3] = 0
        theta[3] = 1

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        r""" Matrix associated with the 2D rigid isoscaling transform

        Parameters
        ----------
        theta : array, shape (4,)
            the parameters of the 2D rigid isoscaling transform
            theta[0] : rotation angle
            theta[1] : translation along the x axis
            theta[2] : translation along the y axis
            theta[3] : isotropic scaling
        R : array, shape (3, 3)
            buffer in which to write the rigid isoscaling matrix
        """
        cdef:
            double ct = cos(theta[0])
            double st = sin(theta[0])
        R[0, 0], R[0, 1], R[0, 2] = ct*theta[3], -st*theta[3], theta[1]
        R[1, 0], R[1, 1], R[1, 2] = st*theta[3], ct*theta[3], theta[2]
        R[2, 0], R[2, 1], R[2, 2] = 0, 0, 1


cdef class RigidIsoScalingTransform3D(Transform):
    def __init__(self):
        """Rigid isoscaling transform in 3D.

        (rotation + translation + scaling)
        The parameter vector theta of length 7 is interpreted as follows:
        theta[0] : rotation about the x axis
        theta[1] : rotation about the y axis
        theta[2] : rotation about the z axis
        theta[3] : translation along the x axis
        theta[4] : translation along the y axis
        theta[5] : translation along the z axis
        theta[6] : isotropic scaling

        """
        self.dim = 3
        self.number_of_parameters = 7

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        r""" Jacobian matrix of a 3D isoscaling rigid transform.

        (rotation + translation + scaling)

        Parameters
        ----------
        theta : array, shape (7,)
            the parameters of the 3D rigid isoscaling transform
            theta[0] : rotation about the x axis
            theta[1] : rotation about the y axis
            theta[2] : rotation about the z axis
            theta[3] : translation along the x axis
            theta[4] : translation along the y axis
            theta[5] : translation along the z axis
            theta[6] : isotropic scaling
        x : array, shape (3,)
            the point at which to compute the Jacobian
        J : array, shape (3, 7)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 0, indicating that the Jacobian is not
            constant (it depends on the value of x)
        """
        cdef:
            double sa = sin(theta[0])
            double ca = cos(theta[0])
            double sb = sin(theta[1])
            double cb = cos(theta[1])
            double sc = sin(theta[2])
            double cc = cos(theta[2])
            double px = x[0], py = x[1], pz = x[2]
            double scale = theta[6]

        J[0, 0] = (-sc * ca * sb) * px * scale + (sc * sa) * py * scale + \
                  (sc * ca * cb) * pz * scale
        J[1, 0] = (cc * ca * sb) * px * scale + (-cc * sa) * py * scale + \
                  (-cc * ca * cb) * pz  * scale
        J[2, 0] = (sa * sb) * px * scale + ca * py * scale + \
                  (-sa * cb) * pz * scale

        J[0, 1] = (-cc * sb - sc * sa * cb) * px * scale + \
                  (cc * cb - sc * sa * sb) * pz * scale
        J[1, 1] = (-sc * sb + cc * sa * cb) * px * scale + \
                  (sc * cb + cc * sa * sb) * pz * scale
        J[2, 1] = (-ca * cb) * px * scale + (-ca * sb) * pz * scale

        J[0, 2] = (-sc * cb - cc * sa * sb) * px * scale + \
                  (-cc * ca) * py  * scale + \
                  (-sc * sb + cc * sa * cb) * pz * scale
        J[1, 2] = (cc * cb - sc * sa * sb) * px * scale + \
                  (-sc * ca) * py * scale + \
                  (cc * sb + sc * sa * cb) * pz * scale
        J[2, 2] = 0

        J[0, 3:6] = 0
        J[1, 3:6] = 0
        J[2, 3:6] = 0
        J[0, 3], J[1, 4], J[2, 5] = 1, 1, 1
        J[0, 6] = (cc*cb-sc*sa*sb) * px - sc*ca*py + (cc*sb+sc*sa*cb) * pz
        J[1, 6] = (sc*cb+cc*sa*sb) * px + cc*ca*py + (sc*sb-cc*sa*cb) * pz
        J[2, 6] = -ca*sb*px + sa*py + ca*cb*pz
        # This Jacobian depends on x (it's not constant): return 0
        return 0

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        """ Parameter values corresponding to the identity

        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (7,)
            buffer to write the parameters of the 3D rigid isoscaling transform
            theta[0] : rotation about the x axis
            theta[1] : rotation about the y axis
            theta[2] : rotation about the z axis
            theta[3] : translation along the x axis
            theta[4] : translation along the y axis
            theta[5] : translation along the z axis
            theta[6] : isotropic scaling

        """
        theta[:6] = 0
        theta[6] = 1

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        """ Matrix associated with the 3D rigid isoscaling transform

        Parameters
        ----------
        theta : array, shape (7,)
            the parameters of the 3D rigid isoscaling transform
            theta[0] : rotation about the x axis
            theta[1] : rotation about the y axis
            theta[2] : rotation about the z axis
            theta[3] : translation along the x axis
            theta[4] : translation along the y axis
            theta[5] : translation along the z axis
            theta[6] : isotropic scaling
        R : array, shape (4, 4)
            buffer in which to write the rigid isoscaling matrix
        """
        cdef:
            double sa = sin(theta[0])
            double ca = cos(theta[0])
            double sb = sin(theta[1])
            double cb = cos(theta[1])
            double sc = sin(theta[2])
            double cc = cos(theta[2])
            double dx = theta[3]
            double dy = theta[4]
            double dz = theta[5]
            double sxyz = theta[6]

        R[0,0], R[0,1], R[0,2] = (cc*cb-sc*sa*sb)*sxyz, -sc*ca*sxyz, (cc*sb+sc*sa*cb)*sxyz
        R[1,0], R[1,1], R[1,2] = (sc*cb+cc*sa*sb)*sxyz, cc*ca*sxyz, (sc*sb-cc*sa*cb)*sxyz
        R[2,0], R[2,1], R[2,2] = -ca*sb*sxyz, sa*sxyz, ca*cb*sxyz
        R[3,0], R[3,1], R[3,2] = 0, 0, 0
        R[0,3] = dx
        R[1,3] = dy
        R[2,3] = dz
        R[3,3] = 1


cdef class RigidScalingTransform2D(Transform):
    def __init__(self):
        """ Rigid scaling transform in 2D.

        (rotation + translation + scaling)

        The parameter vector theta of length 5 is interpreted as follows:
        theta[0] : rotation angle
        theta[1] : translation along the x axis
        theta[2] : translation along the y axis
        theta[3] : scaling along the x axis
        theta[4] : scaling along the y axis
        """
        self.dim = 2
        self.number_of_parameters = 5

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        r""" Jacobian matrix of a 2D rigid scaling transform.

        The transformation (rotation + translation + scaling) is given by:

        T(x,y) = (T1(x,y), T2(x,y)) =
            (sx * x * cost -  sx * y * sint + dx, sy * x * sint + sy * y * cost + dy)

        Parameters
        ----------
        theta : array, shape (5,)
            the parameters of the 2D rigid isoscaling transform
            theta[0] : rotation angle (t)
            theta[1] : translation along the x axis (dx)
            theta[2] : translation along the y axis (dy)
            theta[3] : scaling along the x axis
            theta[4] : scaling along the y axis
        x : array, shape (2,)
            the point at which to compute the Jacobian
        J : array, shape (2, 5)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 0, indicating that the Jacobian is not
            constant (it depends on the value of x)
        """
        cdef:
            double st = sin(theta[0])
            double ct = cos(theta[0])
            double px = x[0], py = x[1]
            double fx = theta[3]
            double fy = theta[4]

        J[0, 0], J[0, 1], J[0, 2] = -px * fx * st - py * fx * ct, 1, 0
        J[1, 0], J[1, 1], J[1, 2] = px * fy * ct - py * fy * st, 0, 1

        J[0, 3], J[0, 4] = px * ct - py * st, 0
        J[1, 3], J[1, 4] = 0, px * st + py * ct
        # This Jacobian depends on x (it's not constant): return 0
        return 0

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        """ Parameter values corresponding to the identity
        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (5,)
            buffer to write the parameters of the 2D scaling rigid transform
            theta[0] : rotation angle
            theta[1] : translation along the x axis
            theta[2] : translation along the y axis
            theta[3] : scaling along the x axis
            theta[4] : scaling along the y axis
        """
        theta[:3] = 0
        theta[3], theta[4] = 1, 1

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        r""" Matrix associated with the 2D rigid scaling transform

        Parameters
        ----------
        theta : array, shape (5,)
            the parameters of the 2D rigid scaling transform
            theta[0] : rotation angle
            theta[1] : translation along the x axis
            theta[2] : translation along the y axis
            theta[3] : scaling along the x axis
            theta[4] : scaling along the y axis
        R : array, shape (3, 3)
            buffer in which to write the rigid scaling matrix
        """
        cdef:
            double ct = cos(theta[0])
            double st = sin(theta[0])
        R[0, 0], R[0, 1], R[0, 2] = ct*theta[3], -st*theta[3], theta[1]
        R[1, 0], R[1, 1], R[1, 2] = st*theta[4], ct*theta[4], theta[2]
        R[2, 0], R[2, 1], R[2, 2] = 0, 0, 1


cdef class RigidScalingTransform3D(Transform):
    def __init__(self):
        """Rigid Scaling transform in 3D (rotation + translation + scaling).

        The parameter vector theta of length 9 is interpreted as follows:
        theta[0] : rotation about the x axis
        theta[1] : rotation about the y axis
        theta[2] : rotation about the z axis
        theta[3] : translation along the x axis
        theta[4] : translation along the y axis
        theta[5] : translation along the z axis
        theta[6] : scaling in the x axis
        theta[7] : scaling in the y axis
        theta[8] : scaling in the z axis

        """
        self.dim = 3
        self.number_of_parameters = 9

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        """Jacobian matrix of a 3D rigid scaling transform.

        (rotation + translation + scaling)

        Parameters
        ----------
        theta : array, shape (9,)
            the parameters of the 3D rigid transform
            theta[0] : rotation about the x axis
            theta[1] : rotation about the y axis
            theta[2] : rotation about the z axis
            theta[3] : translation along the x axis
            theta[4] : translation along the y axis
            theta[5] : translation along the z axis
            theta[6] : scaling in the x axis
            theta[7] : scaling in the y axis
            theta[8] : scaling in the z axis
        x : array, shape (3,)
            the point at which to compute the Jacobian
        J : array, shape (3, 9)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 0, indicating that the Jacobian is not
            constant (it depends on the value of x)
        """
        cdef:
            double sa = sin(theta[0])
            double ca = cos(theta[0])
            double sb = sin(theta[1])
            double cb = cos(theta[1])
            double sc = sin(theta[2])
            double cc = cos(theta[2])
            double px = x[0], py = x[1], pz = x[2]
            double fx = theta[6]
            double fy = theta[7]
            double fz = theta[8]

        J[0, 0] = (-sc * ca * sb) * px * fx + (sc * sa) * py * fx + \
                  (sc * ca * cb) * pz * fx
        J[1, 0] = (cc * ca * sb) * px * fy + (-cc * sa) * py * fy + \
                  (-cc * ca * cb) * pz * fy
        J[2, 0] = (sa * sb) * px * fz + ca * py * fz + (-sa * cb) * pz * fz

        J[0, 1] = (-cc * sb - sc * sa * cb) * px * fx + \
                  (cc * cb - sc * sa * sb) * pz * fx
        J[1, 1] = (-sc * sb + cc * sa * cb) * px * fy + \
                  (sc * cb + cc * sa * sb) * pz * fy
        J[2, 1] = (-ca * cb) * px * fz + (-ca * sb) * pz * fz

        J[0, 2] = (-sc * cb - cc * sa * sb) * px * fx + (-cc * ca) * py * fx + \
                  (-sc * sb + cc * sa * cb) * pz * fx
        J[1, 2] = (cc * cb - sc * sa * sb) * px * fy + (-sc * ca) * py * fy + \
                  (cc * sb + sc * sa * cb) * pz * fy
        J[2, 2] = 0

        J[0, 3:6] = 0
        J[1, 3:6] = 0
        J[2, 3:6] = 0
        J[0, 3], J[1, 4], J[2, 5] = 1, 1, 1
        J[0, 6] = (cc*cb-sc*sa*sb) * px - sc*ca*py + (cc*sb+sc*sa*cb) * pz
        J[1, 6], J[2, 6] = 0, 0
        J[1, 7] = (sc*cb+cc*sa*sb) * px + cc*ca*py + (sc*sb-cc*sa*cb) * pz
        J[0, 7], J[2, 7] = 0, 0
        J[0, 8], J[1, 8] = 0, 0
        J[2, 8] = -ca*sb*px + sa*py + ca*cb*pz
        # This Jacobian depends on x (it's not constant): return 0
        return 0

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        r""" Parameter values corresponding to the identity
        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (9,)
            buffer to write the parameters of the 3D rigid scaling transform
            theta[0] : rotation about the x axis
            theta[1] : rotation about the y axis
            theta[2] : rotation about the z axis
            theta[3] : translation along the x axis
            theta[4] : translation along the y axis
            theta[5] : translation along the z axis
            theta[6] : scaling in the x axis
            theta[7] : scaling in the y axis
            theta[8] : scaling in the z axis
        """
        theta[:6] = 0
        theta[6:9] = 1

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        r""" Matrix associated with the 3D rigid scaling transform

        Parameters
        ----------
        theta : array, shape (9,)
            the parameters of the 3D rigid scaling transform
            theta[0] : rotation about the x axis
            theta[1] : rotation about the y axis
            theta[2] : rotation about the z axis
            theta[3] : translation along the x axis
            theta[4] : translation along the y axis
            theta[5] : translation along the z axis
            theta[6] : scaling in the x axis
            theta[7] : scaling in the y axis
            theta[8] : scaling in the z axis
        R : array, shape (4, 4)
            buffer in which to write the rigid matrix
        """
        cdef:
            double sa = sin(theta[0])
            double ca = cos(theta[0])
            double sb = sin(theta[1])
            double cb = cos(theta[1])
            double sc = sin(theta[2])
            double cc = cos(theta[2])
            double dx = theta[3]
            double dy = theta[4]
            double dz = theta[5]
            double fx = theta[6]
            double fy = theta[7]
            double fz = theta[8]

        R[0,0], R[0,1], R[0,2] = (cc*cb-sc*sa*sb)*fx, -sc*ca*fx, (cc*sb+sc*sa*cb)*fx
        R[1,0], R[1,1], R[1,2] = (sc*cb+cc*sa*sb)*fy, cc*ca*fy, (sc*sb-cc*sa*cb)*fy
        R[2,0], R[2,1], R[2,2] = -ca*sb*fz, sa*fz, ca*cb*fz
        R[3,0], R[3,1], R[3,2] = 0, 0, 0
        R[0,3] = dx
        R[1,3] = dy
        R[2,3] = dz
        R[3,3] = 1


cdef class ScalingTransform2D(Transform):
    def __init__(self):
        r""" Scaling transform in 2D
        """
        self.dim = 2
        self.number_of_parameters = 1

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        r""" Jacobian matrix of the isotropic 2D scale transform
        The transformation is given by:

        T(x) = (s*x0, s*x1)

        The derivative w.r.t. s is T'(x) = [x0, x1]

        Parameters
        ----------
        theta : array, shape (1,)
            the scale factor (the Jacobian does not depend on the scale factor,
            but we receive the buffer to make it consistent with other Jacobian
            functions)
        x : array, shape (2,)
            the point at which to compute the Jacobian
        J : array, shape (2, 1)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 0, indicating that the Jacobian is not
            constant (it depends on the value of x)
        """
        J[0, 0], J[1, 0] = x[0], x[1]
        # This Jacobian depends on x (it's not constant): return 0
        return 0

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        r""" Parameter values corresponding to the identity
        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (1,)
            buffer to write the parameters of the 2D scale transform
        """
        theta[0] = 1

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        r""" Matrix associated with the 2D (isotropic) scaling transform

        Parameters
        ----------
        theta : array, shape (1,)
            the scale factor
        R : array, shape (3, 3)
            the buffer in which to write the scaling matrix
        """
        R[0, 0], R[0, 1], R[0, 2] = theta[0], 0, 0
        R[1, 0], R[1, 1], R[1, 2] = 0, theta[0], 0
        R[2, 0], R[2, 1], R[2, 2] = 0, 0, 1


cdef class ScalingTransform3D(Transform):
    def __init__(self):
        r""" Scaling transform in 3D
        """
        self.dim = 3
        self.number_of_parameters = 1

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        r""" Jacobian matrix of the isotropic 3D scale transform
        The transformation is given by:

        T(x) = (s*x0, s*x1, s*x2)

        The derivative w.r.t. s is T'(x) = [x0, x1, x2]

        Parameters
        ----------
        theta : array, shape (1,)
            the scale factor (the Jacobian does not depend on the scale factor,
            but we receive the buffer to make it consistent with other Jacobian
            functions)
        x : array, shape (3,)
            the point at which to compute the Jacobian
        J : array, shape (3, 1)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 0, indicating that the Jacobian is not
            constant (it depends on the value of x)
        """
        J[0, 0], J[1, 0], J[2, 0]= x[0], x[1], x[2]
        # This Jacobian depends on x (it's not constant): return 0
        return 0

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        r""" Parameter values corresponding to the identity
        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (1,)
            buffer to write the parameters of the 3D scale transform
        """
        theta[0] = 1

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        r""" Matrix associated with the 3D (isotropic) scaling transform

        Parameters
        ----------
        theta : array, shape (1,)
            the scale factor
        R : array, shape (4, 4)
            the buffer in which to write the scaling matrix
        """
        R[0, 0], R[0, 1], R[0, 2], R[0, 3] = theta[0], 0, 0, 0
        R[1, 0], R[1, 1], R[1, 2], R[1, 3] = 0, theta[0], 0, 0
        R[2, 0], R[2, 1], R[2, 2], R[2, 3] = 0, 0, theta[0], 0
        R[3, 0], R[3, 1], R[3, 2], R[3, 3] = 0, 0, 0, 1


cdef class AffineTransform2D(Transform):
    def __init__(self):
        r""" Affine transform in 2D
        """
        self.dim = 2
        self.number_of_parameters = 6

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        r""" Jacobian matrix of the 2D affine transform
        The transformation is given by:

        T(x) = |a0, a1, a2 |   |x0|   | T1(x) |   |a0*x0 + a1*x1 + a2|
               |a3, a4, a5 | * |x1| = | T2(x) | = |a3*x0 + a4*x1 + a5|
                               | 1|

        The derivatives w.r.t. each parameter are given by

        T'(x) = [[x0,  0], #derivatives of [T1, T2] w.r.t a0
                 [x1,  0], #derivatives of [T1, T2] w.r.t a1
                 [ 1,  0], #derivatives of [T1, T2] w.r.t a2
                 [ 0, x0], #derivatives of [T1, T2] w.r.t a3
                 [ 0, x1], #derivatives of [T1, T2] w.r.t a4
                 [ 0,  1]] #derivatives of [T1, T2, T3] w.r.t a5

        The Jacobian matrix is the transpose of the above matrix.

        Parameters
        ----------
        theta : array, shape (6,)
            the parameters of the 2D affine transform
        x : array, shape (2,)
            the point at which to compute the Jacobian
        J : array, shape (2, 6)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 0, indicating that the Jacobian is not
            constant (it depends on the value of x)
        """
        J[0, :6] = 0
        J[1, :6] = 0

        J[0, :2] = x[:2]
        J[0, 2] = 1
        J[1, 3:5] = x[:2]
        J[1, 5] = 1
        # This Jacobian depends on x (it's not constant): return 0
        return 0

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        r""" Parameter values corresponding to the identity
        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (6,)
            buffer to write the parameters of the 2D affine transform
        """
        theta[0], theta[1], theta[2] = 1, 0, 0
        theta[3], theta[4], theta[5] = 0, 1, 0

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        r""" Matrix associated with a general 2D affine transform

        The transformation is given by the matrix:

        A = [[a0, a1, a2],
             [a3, a4, a5],
             [ 0,  0,  1]]

        Parameters
        ----------
        theta : array, shape (6,)
            the parameters of the 2D affine transform
        R : array, shape (3,3)
            the buffer in which to write the matrix
        """
        R[0, 0], R[0, 1], R[0, 2] = theta[0], theta[1], theta[2]
        R[1, 0], R[1, 1], R[1, 2] = theta[3], theta[4], theta[5]
        R[2, 0], R[2, 1], R[2, 2] = 0, 0, 1


cdef class AffineTransform3D(Transform):
    def __init__(self):
        r""" Affine transform in 3D
        """
        self.dim = 3
        self.number_of_parameters = 12

    cdef int _jacobian(self, double[:] theta, double[:] x,
                       double[:, :] J) noexcept nogil:
        r""" Jacobian matrix of the 3D affine transform
        The transformation is given by:

        T(x)= |a0, a1, a2,  a3 |  |x0|  | T1(x) |  |a0*x0 + a1*x1 + a2*x2 + a3|
              |a4, a5, a6,  a7 |* |x1|= | T2(x) |= |a4*x0 + a5*x1 + a6*x2 + a7|
              |a8, a9, a10, a11|  |x2|  | T3(x) |  |a8*x0 + a9*x1 + a10*x2+a11|
                                  | 1|

        The derivatives w.r.t. each parameter are given by

        T'(x) = [[x0,  0,  0], #derivatives of [T1, T2, T3] w.r.t a0
                 [x1,  0,  0], #derivatives of [T1, T2, T3] w.r.t a1
                 [x2,  0,  0], #derivatives of [T1, T2, T3] w.r.t a2
                 [ 1,  0,  0], #derivatives of [T1, T2, T3] w.r.t a3
                 [ 0, x0,  0], #derivatives of [T1, T2, T3] w.r.t a4
                 [ 0, x1,  0], #derivatives of [T1, T2, T3] w.r.t a5
                 [ 0, x2,  0], #derivatives of [T1, T2, T3] w.r.t a6
                 [ 0,  1,  0], #derivatives of [T1, T2, T3] w.r.t a7
                 [ 0,  0, x0], #derivatives of [T1, T2, T3] w.r.t a8
                 [ 0,  0, x1], #derivatives of [T1, T2, T3] w.r.t a9
                 [ 0,  0, x2], #derivatives of [T1, T2, T3] w.r.t a10
                 [ 0,  0,  1]] #derivatives of [T1, T2, T3] w.r.t a11

        The Jacobian matrix is the transpose of the above matrix.

        Parameters
        ----------
        theta : array, shape (12,)
            the parameters of the 3D affine transform
        x : array, shape (3,)
            the point at which to compute the Jacobian
        J : array, shape (3, 12)
            the buffer in which to write the Jacobian

        Returns
        -------
        is_constant : int
            always returns 0, indicating that the Jacobian is not
            constant (it depends on the value of x)
        """
        cdef:
            cnp.npy_intp j

        for j in range(3):
            J[j, :12] = 0
        J[0, :3] = x[:3]
        J[0, 3] = 1
        J[1, 4:7] = x[:3]
        J[1, 7] = 1
        J[2, 8:11] = x[:3]
        J[2, 11] = 1
        # This Jacobian depends on x (it's not constant): return 0
        return 0

    cdef void _get_identity_parameters(self, double[:] theta) noexcept nogil:
        r""" Parameter values corresponding to the identity
        Sets in theta the parameter values corresponding to the identity
        transform

        Parameters
        ----------
        theta : array, shape (12,)
            buffer to write the parameters of the 3D affine transform
        """
        theta[0], theta[1], theta[2], theta[3] = 1, 0, 0, 0
        theta[4], theta[5], theta[6], theta[7] = 0, 1, 0, 0
        theta[8], theta[9], theta[10], theta[11] = 0, 0, 1, 0

    cdef void _param_to_matrix(self, double[:] theta, double[:, :] R) noexcept nogil:
        r""" Matrix associated with a general 3D affine transform

        The transformation is given by the matrix:

        A = [[a0, a1, a2, a3],
             [a4, a5, a6, a7],
             [a8, a9, a10, a11],
             [ 0,   0,   0,   1]]

        Parameters
        ----------
        theta : array, shape (12,)
            the parameters of the 3D affine transform
        R : array, shape (4,4)
            the buffer in which to write the matrix
        """
        R[0, 0], R[0, 1], R[0, 2] = theta[0], theta[1], theta[2]
        R[1, 0], R[1, 1], R[1, 2] = theta[4], theta[5], theta[6]
        R[2, 0], R[2, 1], R[2, 2] = theta[8], theta[9], theta[10]
        R[3, 0], R[3, 1], R[3, 2] = 0, 0, 0
        R[0, 3] = theta[3]
        R[1, 3] = theta[7]
        R[2, 3] = theta[11]
        R[3, 3] = 1


regtransforms = dict()
regtransforms [('TRANSLATION', 2)] = TranslationTransform2D()
regtransforms [('TRANSLATION', 3)] = TranslationTransform3D()
regtransforms [('ROTATION', 2)] = RotationTransform2D()
regtransforms [('ROTATION', 3)] = RotationTransform3D()
regtransforms [('RIGID', 2)] = RigidTransform2D()
regtransforms [('RIGID', 3)] = RigidTransform3D()
regtransforms [('SCALING', 2)] = ScalingTransform2D()
regtransforms [('SCALING', 3)] = ScalingTransform3D()
regtransforms [('AFFINE', 2)] = AffineTransform2D()
regtransforms [('AFFINE', 3)] = AffineTransform3D()
regtransforms [('RIGIDSCALING', 2)] = RigidScalingTransform2D()
regtransforms [('RIGIDSCALING', 3)] = RigidScalingTransform3D()
regtransforms [('RIGIDISOSCALING', 2)] = RigidIsoScalingTransform2D()
regtransforms [('RIGIDISOSCALING', 3)] = RigidIsoScalingTransform3D()



from functools import partial
import multiprocessing as mp

import numpy as np
import scipy.fft

from dipy.testing.decorators import warning_for_keywords
from dipy.utils.multiproc import determine_num_processes

_fft = scipy.fft


@warning_for_keywords()
def _image_tv(x, *, axis=0, n_points=3):
    """Computes total variation (TV) of matrix x across a given axis and
    along two directions.

    Parameters
    ----------
    x : 2D ndarray
        matrix x
    axis : int (0 or 1)
        Axis which TV will be calculated. Default a is set to 0.
    n_points : int
        Number of points to be included in TV calculation.

    Returns
    -------
    ptv : 2D ndarray
        Total variation calculated from the right neighbours of each point.
    ntv : 2D ndarray
        Total variation calculated from the left neighbours of each point.

    """
    xs = x.copy() if axis else x.T.copy()

    # Add copies of the data so that data extreme points are also analysed
    xs = np.concatenate(
        (xs[:, (-n_points - 1) :], xs, xs[:, 0 : (n_points + 1)]), axis=1
    )

    ptv = np.absolute(
        xs[:, (n_points + 1) : (-n_points - 1)] - xs[:, (n_points + 2) : (-n_points)]
    )
    ntv = np.absolute(
        xs[:, (n_points + 1) : (-n_points - 1)] - xs[:, n_points : (-n_points - 2)]
    )
    for n in range(1, n_points):
        ptv = ptv + np.absolute(
            xs[:, (n_points + 1 + n) : (-n_points - 1 + n)]
            - xs[:, (n_points + 2 + n) : (-n_points + n)]
        )
        ntv = ntv + np.absolute(
            xs[:, (n_points + 1 - n) : (-n_points - 1 - n)]
            - xs[:, (n_points - n) : (-n_points - 2 - n)]
        )

    if axis:
        return ptv, ntv
    else:
        return ptv.T, ntv.T


@warning_for_keywords()
def _gibbs_removal_1d(x, *, axis=0, n_points=3):
    """Suppresses Gibbs ringing along a given axis using fourier sub-shifts.

    Parameters
    ----------
    x : 2D ndarray
        Matrix x.
    axis : int (0 or 1)
        Axis in which Gibbs oscillations will be suppressed.
        Default is set to 0.
    n_points : int, optional
        Number of neighbours to access local TV (see note).
        Default is set to 3.

    Returns
    -------
    xc : 2D ndarray
        Matrix with suppressed Gibbs oscillations along the given axis.

    Notes
    -----
    This function suppresses the effects of Gibbs oscillations based on the
    analysis of local total variation (TV). Although artefact correction is
    done based on two adjacent points for each voxel, total variation should be
    accessed in a larger range of neighbours. The number of neighbours to be
    considered in TV calculation can be adjusted using the parameter n_points.

    """
    dtype_float = np.promote_types(x.real.dtype, np.float32)

    ssamp = np.linspace(0.02, 0.9, num=45, dtype=dtype_float)

    xs = x.copy() if axis else x.T.copy()

    # TV for shift zero (baseline)
    tvr, tvl = _image_tv(xs, axis=1, n_points=n_points)
    tvp = np.minimum(tvr, tvl)
    tvn = tvp.copy()

    # Find optimal shift for gibbs removal
    isp = xs.copy()
    isn = xs.copy()
    sp = np.zeros(xs.shape, dtype=dtype_float)
    sn = np.zeros(xs.shape, dtype=dtype_float)
    N = xs.shape[1]
    c = _fft.fft(xs, axis=1)
    k = _fft.fftfreq(N, 1 / (2.0j * np.pi))
    k = k.astype(c.dtype, copy=False)
    for s in ssamp:
        ks = k * s
        # Access positive shift for given s
        img_p = abs(_fft.ifft(c * np.exp(ks), axis=1))

        tvsr, tvsl = _image_tv(img_p, axis=1, n_points=n_points)
        tvs_p = np.minimum(tvsr, tvsl)

        # Access negative shift for given s
        img_n = abs(_fft.ifft(c * np.exp(-ks), axis=1))
        tvsr, tvsl = _image_tv(img_n, axis=1, n_points=n_points)
        tvs_n = np.minimum(tvsr, tvsl)

        # Update positive shift params
        isp[tvp > tvs_p] = img_p[tvp > tvs_p]
        sp[tvp > tvs_p] = s
        tvp[tvp > tvs_p] = tvs_p[tvp > tvs_p]

        # Update negative shift params
        isn[tvn > tvs_n] = img_n[tvn > tvs_n]
        sn[tvn > tvs_n] = s
        tvn[tvn > tvs_n] = tvs_n[tvn > tvs_n]

    # check non-zero sub-voxel shifts
    idx = np.nonzero(sp + sn)

    # use positive and negative optimal sub-voxel shifts to interpolate to
    # original grid points
    xs[idx] = (isp[idx] - isn[idx]) / (sp[idx] + sn[idx]) * sn[idx] + isn[idx]

    return xs if axis else xs.T


def _weights(shape):
    """Computes the weights necessary to combine two images processed by
    the 1D Gibbs removal procedure along two different axes.

    See :footcite:p:`Kellner2016` for further details about the method.

    Parameters
    ----------
    shape : tuple
        shape of the image.

    Returns
    -------
    G0 : 2D ndarray
        Weights for the image corrected along axis 0.
    G1 : 2D ndarray
        Weights for the image corrected along axis 1.

    References
    ----------
    .. footbibliography::

    """
    G0 = np.zeros(shape)
    G1 = np.zeros(shape)
    k0 = np.linspace(-np.pi, np.pi, num=shape[0])
    k1 = np.linspace(-np.pi, np.pi, num=shape[1])

    # Middle points
    K1, K0 = np.meshgrid(k1[1:-1], k0[1:-1])
    cosk0 = 1.0 + np.cos(K0)
    cosk1 = 1.0 + np.cos(K1)
    G1[1:-1, 1:-1] = cosk0 / (cosk0 + cosk1)
    G0[1:-1, 1:-1] = cosk1 / (cosk0 + cosk1)

    # Boundaries
    G1[1:-1, 0] = G1[1:-1, -1] = 1
    G1[0, 0] = G1[-1, -1] = G1[0, -1] = G1[-1, 0] = 1 / 2
    G0[0, 1:-1] = G0[-1, 1:-1] = 1
    G0[0, 0] = G0[-1, -1] = G0[0, -1] = G0[-1, 0] = 1 / 2

    return G0, G1


@warning_for_keywords()
def _gibbs_removal_2d(image, *, n_points=3, G0=None, G1=None):
    """Suppress Gibbs ringing of a 2D image :footcite:p:`Kellner2016`,
    :footcite:p:`NetoHenriques2018`.

    Parameters
    ----------
    image : 2D ndarray
        Matrix containing the 2D image.
    n_points : int, optional
        Number of neighbours to access local TV (see note). Default is
        set to 3.
    G0 : 2D ndarray, optional
        Weights for the image corrected along axis 0. If not given, the
        function estimates them using the function :func:`_weights`.
    G1 : 2D ndarray, optional
        Weights for the image corrected along axis 1. If not given, the
        function estimates them using the function :func:`_weights`.

    Returns
    -------
    imagec : 2D ndarray
        Matrix with Gibbs oscillations reduced along axis a.

    Notes
    -----
    This function suppresses the effects of Gibbs oscillations based on the
    analysis of local total variation (TV). Although artefact correction is
    done based on two adjacent points for each voxel, total variation should be
    accessed in a larger range of neighbours. The number of neighbours to be
    considered in TV calculation can be adjusted using the parameter n_points.

    References
    ----------
    .. footbibliography::

    """
    if G0 is None or G1 is None:
        G0, G1 = _weights(image.shape)

    img_c1 = _gibbs_removal_1d(image, axis=1, n_points=n_points)
    img_c0 = _gibbs_removal_1d(image, axis=0, n_points=n_points)

    C1 = _fft.fft2(img_c1)
    C0 = _fft.fft2(img_c0)
    imagec = abs(_fft.ifft2(_fft.fftshift(C1) * G1 + _fft.fftshift(C0) * G0))

    return imagec


@warning_for_keywords()
def gibbs_removal(vol, *, slice_axis=2, n_points=3, inplace=True, num_processes=1):
    """Suppresses Gibbs ringing artefacts of images volumes.

    See :footcite:p:`Kellner2016` and :footcite:p:`NetoHenriques2018` for
    further details about the method.

    Parameters
    ----------
    vol : ndarray ([X, Y]), ([X, Y, Z]) or ([X, Y, Z, g])
        Matrix containing one volume (3D) or multiple (4D) volumes of images.
    slice_axis : int (0, 1, or 2)
        Data axis corresponding to the number of acquired slices.
    n_points : int, optional
        Number of neighbour points to access local TV (see note).
    inplace : bool, optional
        If True, the input data is replaced with results. Otherwise, returns
        a new array.
    num_processes : int or None, optional
        Split the calculation to a pool of children processes. This only
        applies to 3D or 4D `data` arrays. Default is 1. If < 0 the maximal
        number of cores minus ``num_processes + 1`` is used (enter -1 to use
        as many cores as possible). 0 raises an error.

    Returns
    -------
    vol : ndarray ([X, Y]), ([X, Y, Z]) or ([X, Y, Z, g])
        Matrix containing one volume (3D) or multiple (4D) volumes of corrected
        images.

    Notes
    -----
    For 4D matrix last element should always correspond to the number of
    diffusion gradient directions.

    References
    ----------
    .. footbibliography::

    """
    nd = vol.ndim

    # check matrix dimension
    if nd > 4:
        raise ValueError("Data have to be a 4D, 3D or 2D matrix")
    elif nd < 2:
        raise ValueError("Data is not an image")

    if not isinstance(inplace, bool):
        raise TypeError("inplace must be a boolean.")

    num_processes = determine_num_processes(num_processes)

    # check the axis corresponding to different slices
    # 1) This axis cannot be larger than 2
    if slice_axis > 2:
        raise ValueError(
            "Different slices have to be organized along"
            + "one of the 3 first matrix dimensions"
        )

    # 2) Reorder axis to allow iteration over the first axis
    elif nd == 3:
        vol = np.moveaxis(vol, slice_axis, 0)
    elif nd == 4:
        vol = np.moveaxis(vol, (slice_axis, 3), (0, 1))

    if nd == 4:
        inishap = vol.shape
        vol = vol.reshape((inishap[0] * inishap[1], inishap[2], inishap[3]))

    # Produce weighting functions for 2D Gibbs removal
    shap = vol.shape
    G0, G1 = _weights(shap[-2:])

    # Copy data if not inplace
    if not inplace:
        vol = vol.copy()

    # Run Gibbs removal of 2D images
    if nd == 2:
        vol[:, :] = _gibbs_removal_2d(vol, n_points=n_points, G0=G0, G1=G1)
    else:
        if num_processes == 1:
            for i in range(shap[0]):
                vol[i, :, :] = _gibbs_removal_2d(
                    vol[i, :, :], n_points=n_points, G0=G0, G1=G1
                )
        else:
            mp.set_start_method("spawn", force=True)
            pool = mp.Pool(num_processes)

            partial_func = partial(_gibbs_removal_2d, n_points=n_points, G0=G0, G1=G1)
            vol[:, :, :] = pool.map(partial_func, vol)
            pool.close()
            pool.join()

    # Reshape data to original format
    if nd == 3:
        vol = np.moveaxis(vol, 0, slice_axis)
    if nd == 4:
        vol = vol.reshape(inishap)
        vol = np.moveaxis(vol, (0, 1), (slice_axis, 3))

    return vol



cimport cython
cimport numpy as cnp


cdef void remove_similar_vertices_c(
    double[:, :] vertices,
    double theta,
    int remove_antipodal,
    int return_mapping,
    int return_index,
    double[:, :] unique_vertices,
    cnp.uint16_t[:] mapping,
    cnp.uint16_t[:] index,
    cnp.uint16_t* n_unique
) noexcept nogil


cdef cnp.npy_intp search_descending_c(
    cython.floating* arr,
    cnp.npy_intp size,
    double relative_threshold) noexcept nogil


cdef long local_maxima_c(
    double[:] odf, cnp.uint16_t[:, :] edges,
    double[::1] out_values,
    cnp.npy_intp[::1] out_indices) noexcept nogil


"""
Cython implementation of (parts of) the standard library time module.

Note: On implementations that lack a C-API for monotonic/perfcounter clocks
(like PyPy), the fallback code uses the system clock which may return absolute
time values from a different value range, differing from those provided by
Python's "time" module.

Note: To remove and use Cython when the cython minimum version of the
project will be 3.1.0
"""

from libc.stdint cimport int64_t
from cpython.exc cimport PyErr_SetFromErrno

cdef extern from *:
    """
    #if PY_VERSION_HEX >= 0x030d00b1 || defined(PyTime_t)
        #define __Pyx_PyTime_t PyTime_t
    #else
        #define __Pyx_PyTime_t _PyTime_t
    #endif

    #if PY_VERSION_HEX >= 0x030d00b1 || defined(PyTime_TimeRaw)
        static CYTHON_INLINE __Pyx_PyTime_t __Pyx_PyTime_TimeRaw(void) {
            __Pyx_PyTime_t tic;
            (void) PyTime_TimeRaw(&tic);
            return tic;
        }
    #else
        #define __Pyx_PyTime_TimeRaw()  _PyTime_GetSystemClock()
    #endif

    #if PY_VERSION_HEX >= 0x030d00b1 || defined(PyTime_MonotonicRaw)
        static CYTHON_INLINE __Pyx_PyTime_t __Pyx_PyTime_MonotonicRaw(void) {
            __Pyx_PyTime_t tic;
            (void) PyTime_MonotonicRaw(&tic);
            return tic;
        }
    #elif CYTHON_COMPILING_IN_PYPY && !defined(_PyTime_GetMonotonicClock)
        #define __Pyx_PyTime_MonotonicRaw()  _PyTime_GetSystemClock()
    #else
        #define __Pyx_PyTime_MonotonicRaw()  _PyTime_GetMonotonicClock()
    #endif

    #if PY_VERSION_HEX >= 0x030d00b1 || defined(PyTime_PerfCounterRaw)
        static CYTHON_INLINE __Pyx_PyTime_t __Pyx_PyTime_PerfCounterRaw(void) {
            __Pyx_PyTime_t tic;
            (void) PyTime_PerfCounterRaw(&tic);
            return tic;
        }
    #elif CYTHON_COMPILING_IN_PYPY && !defined(_PyTime_GetPerfCounter)
        #define __Pyx_PyTime_PerfCounterRaw()  __Pyx_PyTime_MonotonicRaw()
    #else
        #define __Pyx_PyTime_PerfCounterRaw()  _PyTime_GetPerfCounter()
    #endif

    #if PY_VERSION_HEX >= 0x030d00b1 || defined(PyTime_AsSecondsDouble)
        #define __Pyx_PyTime_AsSecondsDouble(t)  PyTime_AsSecondsDouble(t)
    #else
        #define __Pyx_PyTime_AsSecondsDouble(t)  _PyTime_AsSecondsDouble(t)
    #endif
    """
    ctypedef int64_t PyTime_t "__Pyx_PyTime_t"

    ctypedef PyTime_t _PyTime_t "__Pyx_PyTime_t"  # legacy, use "PyTime_t" instead
    PyTime_t PyTime_TimeUnchecked "__Pyx_PyTime_TimeRaw" () noexcept nogil  # legacy, use "PyTime_TimeRaw" instead

    PyTime_t PyTime_TimeRaw "__Pyx_PyTime_TimeRaw" () noexcept nogil
    PyTime_t PyTime_MonotonicRaw "__Pyx_PyTime_MonotonicRaw" () noexcept nogil
    PyTime_t PyTime_PerfCounterRaw "__Pyx_PyTime_PerfCounterRaw" () noexcept nogil
    double PyTime_AsSecondsDouble "__Pyx_PyTime_AsSecondsDouble" (PyTime_t t) noexcept nogil


from libc.time cimport (
    tm,
    time_t,
    localtime as libc_localtime,
)


cdef inline double time() noexcept nogil:
    cdef PyTime_t tic = PyTime_TimeRaw()
    return PyTime_AsSecondsDouble(tic)


cdef inline int64_t time_ns() noexcept nogil:
    return <int64_t> PyTime_TimeRaw()


cdef inline double perf_counter() noexcept nogil:
    cdef PyTime_t tic = PyTime_PerfCounterRaw()
    return PyTime_AsSecondsDouble(tic)


cdef inline int64_t perf_counter_ns() noexcept nogil:
    return <int64_t> PyTime_PerfCounterRaw()


cdef inline double monotonic() noexcept nogil:
    cdef PyTime_t tic = PyTime_MonotonicRaw()
    return PyTime_AsSecondsDouble(tic)


cdef inline int64_t monotonic_ns() noexcept nogil:
    return <int64_t> PyTime_MonotonicRaw()


cdef inline int _raise_from_errno() except -1 with gil:
    PyErr_SetFromErrno(RuntimeError)
    return <int> -1  # Let the C compiler know that this function always raises.


cdef inline tm localtime() except * nogil:
    """
    Analogue to the stdlib time.localtime.  The returned struct
    has some entries that the stdlib version does not: tm_gmtoff, tm_zone
    """
    cdef:
        time_t tic = <time_t>time()
        tm* result

    result = libc_localtime(&tic)
    if result is NULL:
        _raise_from_errno()
    # Fix 0-based date values (and the 1900-based year).
    # See tmtotuple() in https://github.com/python/cpython/blob/master/Modules/timemodule.c
    result.tm_year += 1900
    result.tm_mon += 1
    result.tm_wday = <int> ((result.tm_wday + 6) % 7)
    result.tm_yday += 1
    return result[0]


"""
============================================
Nonrigid Bundle Registration with BundleWarp
============================================

This example explains how you can nonlinearly register two bundles from two
different subjects directly in the space of streamlines
:footcite:p:`Chandio2023`, :footcite:p:`Chandio2020b`.

To show the concept, we will use two pre-saved uncinate fasciculus bundles. The
algorithm used here is called BundleWarp, streamline-based nonlinear
registration of white matter tracts :footcite:p:`Chandio2023`.

"""

from os.path import join as pjoin
from time import time

from dipy.align.streamwarp import (
    bundlewarp,
    bundlewarp_shape_analysis,
    bundlewarp_vector_filed,
)
from dipy.data import fetch_bundle_warp_dataset
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import load_trk, save_tractogram
from dipy.tracking.streamline import (
    Streamlines,
    set_number_of_points,
    unlist_streamlines,
)
from dipy.viz.streamline import viz_displacement_mag, viz_two_bundles, viz_vector_field

###############################################################################
# Let's download and load two uncinate fasciculus bundles in the left
# hemisphere of the brain (UF_L) available here:
# https://figshare.com/articles/dataset/Test_Bundles_for_DIPY/22557733

bundle_warp_files = fetch_bundle_warp_dataset()
s_UF_L_path = pjoin(bundle_warp_files[1], "s_UF_L.trk")
m_UF_L_path = pjoin(bundle_warp_files[1], "m_UF_L.trk")

uf_subj1 = load_trk(s_UF_L_path, reference="same", bbox_valid_check=False).streamlines
uf_subj2 = load_trk(m_UF_L_path, reference="same", bbox_valid_check=False).streamlines

###############################################################################
# Let's resample the streamlines so that they both have the same number of
# points per streamline. Here we will use 20 points.

static = Streamlines(set_number_of_points(uf_subj1, nb_points=20))
moving = Streamlines(set_number_of_points(uf_subj2, nb_points=20))

###############################################################################
# We call ``uf_subj2`` a moving bundle as it will be nonlinearly aligned with
# ``uf_subj1`` (static) bundle. Here is how this is done.
#
#
# Let's visualize static bundle in red and moving in green before
# registration.

viz_two_bundles(static, moving, fname="static_and_moving.png")

###############################################################################
# BundleWarp method provides a unique ability to either partially or fully
# deform a moving bundle by the use of a single regularization parameter alpha.
# alpha controls the trade-off between regularizing the deformation and having
# points match very closely. The lower the value of alpha, the more closely
# the bundles would match.
#
# Let's partially deform bundle by setting alpha=0.5.

start = time()
deformed_bundle, moving_aligned, distances, match_pairs, warp_map = bundlewarp(
    static, moving, alpha=0.5, beta=20, max_iter=15
)
end = time()

print("time taken by BundleWarp registration in seconds = ", end - start)

###############################################################################
# Let's visualize static bundle in red and moved (warped) in green. Note: You
# can set interactive=True in visualization functions throughout this tutorial
# if you prefer to get interactive visualization window.

viz_two_bundles(static, deformed_bundle, fname="static_and_partially_deformed.png")

###############################################################################
# Let's visualize linearly moved bundle in blue and nonlinearly moved bundle in
# green to see BundleWarp registration improvement over linear SLR
# registration.

viz_two_bundles(
    moving_aligned,
    deformed_bundle,
    fname="linearly_and_nonlinearly_moved.png",
    c1=(0, 0, 1),
)

###############################################################################
# Now, let's visualize deformation vector field generated by BundleWarp.
# This shows us visually where and how much and in what directions deformations
# were added by BundleWarp.

offsets, directions, colors = bundlewarp_vector_filed(moving_aligned, deformed_bundle)

points_aligned, _ = unlist_streamlines(moving_aligned)

###############################################################################
# Visualizing just the vector field.

fname = "partially_vectorfield.png"
viz_vector_field(points_aligned, directions, colors, offsets, fname)

###############################################################################
# Let's visualize vector field over linearly moved bundle. This will show how
# much deformations were introduced after linear registration.

fname = "partially_vectorfield_over_linearly_moved.png"
viz_vector_field(
    points_aligned, directions, colors, offsets, fname, bundle=moving_aligned
)

###############################################################################
# We can also visualize the magnitude of deformations in mm mapped over
# affinely moved bundle. It shows which streamlines were deformed the most
# after affine registration.

fname = "partially_deformation_magnitude_over_linearly_moved.png"
viz_displacement_mag(moving_aligned, offsets, fname, interactive=False)

###############################################################################
# Saving partially warped bundle.

new_tractogram = StatefulTractogram(deformed_bundle, m_UF_L_path, Space.RASMM)
save_tractogram(new_tractogram, "partially_deformed_bundle.trk", bbox_valid_check=False)


###############################################################################
# Let's fully deform the moving bundle by setting alpha <= 0.01
#
# We will use MDF distances computed and returned by previous run of BundleWarp
# method. This will save computation time.

start = time()
deformed_bundle2, moving_aligned, distances, match_pairs, warp_map = bundlewarp(
    static, moving, dist=distances, alpha=0.001, beta=20
)
end = time()

print("time taken by BundleWarp registration in seconds = ", end - start)

###############################################################################
# Let's visualize static bundle in red and moved (completely warped) in green.

viz_two_bundles(static, deformed_bundle2, fname="static_and_fully_deformed.png")

###############################################################################
# Now, let's visualize the deformation vector field generated by BundleWarp.
# This shows us visually where and how much and in what directions deformations
# were added by BundleWarp to perfectly warp moving bundle to look like static.

offsets, directions, colors = bundlewarp_vector_filed(moving_aligned, deformed_bundle2)

points_aligned, _ = unlist_streamlines(moving_aligned)

###############################################################################
# Visualizing just the vector field.

fname = "fully_vectorfield.png"
viz_vector_field(points_aligned, directions, colors, offsets, fname)

###############################################################################
# Let's visualize vector field over linearly moved bundle. This will show how
# much deformations were introduced after linear registration by fully
# deforming the moving bundle.

fname = "fully_vectorfield_over_linearly_moved.png"
viz_vector_field(
    points_aligned, directions, colors, offsets, fname, bundle=moving_aligned
)

###############################################################################
# Let's visualize the magnitude of deformations in mm mapped over affinely
# moved bundle. It shows which streamlines were deformed the most after affine
# registration.

fname = "fully_deformation_magnitude_over_linearly_moved.png"
viz_displacement_mag(moving_aligned, offsets, fname, interactive=False)

###############################################################################
# We can also perform bundle shape difference analysis using the displacement
# field generated by fully warping the moving bundle to look exactly like
# static bundle. Here, we plot bundle shape profile using BUAN. Bundle shape
# profile shows the average magnitude of deformations along the length of the
# bundle. Segments where we observe higher deformations are the areas where
# two bundles differ the most in shape.

_, _ = bundlewarp_shape_analysis(
    moving_aligned, deformed_bundle, no_disks=10, plotting=False
)

###############################################################################
# Saving fully warped bundle.

new_tractogram = StatefulTractogram(deformed_bundle2, m_UF_L_path, Space.RASMM)
save_tractogram(new_tractogram, "fully_deformed_bundle.trk", bbox_valid_check=False)

###############################################################################
# References
# ----------
#
# .. footbibliography::



from warnings import warn

import nibabel as nib

from dipy.testing.decorators import warning_for_keywords


@warning_for_keywords()
def load_pial(fname, *, return_meta=False):
    """Load pial file.

    Parameters
    ----------
    fname : str
        Absolute path of the file.
    return_meta : bool, optional
        Whether to read the metadata of the file or not, by default False.

    Returns
    -------
    tuple
        (vertices, faces) if return_meta=False. Otherwise, (vertices, faces,
        metadata).
    """
    try:
        return nib.freesurfer.read_geometry(fname, read_metadata=return_meta)
    except ValueError:
        warn(f"The file {fname} provided does not have geometry data.", stacklevel=2)


def load_gifti(fname):
    """Load gifti file.

    Parameters
    ----------
    fname : str
        Absolute path of the file.

    Returns
    -------
    tuple
        (vertices, faces)
    """
    surf_img = nib.load(fname)
    return surf_img.agg_data(("pointset", "triangle"))



#!/usr/bin/python
"""Functions for defining weights for iterative fitting methods."""

import numpy as np

MIN_POSITIVE_SIGNAL = 0.0001


def simple_cutoff(
    residuals, log_residuals, pred_sig, design_matrix, leverages, C, cutoff
):
    """Define outliers based on the signal (rather than the log-signal).

    Parameters
    ----------
    residuals : ndarray
        Residuals of the signal (observed signal - fitted signal).
    log_residuals : ndarray
        Residuals of the log signal (log observed signal - fitted log signal).
    pred_sig : ndarray
        The predicted signal, given a previous fit.
    design_matrix : ndarray (g, ...)
        Design matrix holding the covariants used to solve for the
        regression coefficients.
    leverages : ndarray
        The leverages (diagonal of the 'hat matrix') of the fit.
    C : float
        Estimate of the standard deviation of the error.
    cutoff : float, optional
        Cut-off value for defining outliers based on fitting residuals.
        Here the condition is::

            |residuals| > cut_off x C x HAT_factor

        where HAT_factor = sqrt(1 - leverages) adjusts for leverage effects.

    """
    leverages[np.isclose(leverages, 1.0)] = 0.99  # avoids rare issues
    HAT_factor = np.sqrt(1 - leverages)
    cond = np.abs(residuals) > +cutoff * C * HAT_factor
    return cond


def two_eyes_cutoff(
    residuals, log_residuals, pred_sig, design_matrix, leverages, C, cutoff
):
    """Define outliers with two-eyes approach.

    see :footcite:p:`Collier2015` for more details.

    Parameters
    ----------
    residuals : ndarray
        Residuals of the signal (observed signal - fitted signal).
    log_residuals : ndarray
        Residuals of the log signal (log observed signal - fitted log signal).
    pred_sig : ndarray
        The predicted signal, given a previous fit.
    design_matrix : ndarray (g, ...)
        Design matrix holding the covariants used to solve for the
        regression coefficients.
    leverages : ndarray
        The leverages (diagonal of the 'hat matrix') of the fit.
    C : float
        Estimate of the standard deviation of the error.
    cutoff : float, optional
        Cut-off value for defining outliers based on fitting residuals,
        see :footcite:p:`Collier2015` for the two-eyes approached used here.

    References
    ----------
    .. footbibliography::

    """
    leverages[np.isclose(leverages, 1.0)] = 0.99  # avoids rare issues
    HAT_factor = np.sqrt(1 - leverages)
    cond = (residuals > +cutoff * C * HAT_factor) | (
        log_residuals < -cutoff * C * HAT_factor / pred_sig
    )
    return cond


def weights_method_wls_m_est(
    data,
    pred_sig,
    design_matrix,
    leverages,
    idx,
    total_idx,
    last_robust,
    *,
    m_est="gm",
    cutoff=3,
    outlier_condition_func=simple_cutoff,
):
    """Calculate M-estimator weights for WLS model.

    Parameters
    ----------
    data : ndarray
        The measured signal.
    pred_sig : ndarray
        The predicted signal, given a previous fit.
        Has the same shape as data.
    design_matrix : ndarray (g, ...)
        Design matrix holding the covariants used to solve for the
        regression coefficients.
    leverages : ndarray
        The leverages (diagonal of the 'hat matrix') of the fit.
    idx : int
        The current iteration number.
    total_idx : int
        The total number of iterations.
    last_robust : ndarray
        True for inlier indices and False for outlier indices. Must have the
        same shape as data.
    m_est : str, optional.
        M-estimator weighting scheme to use. Currently,
        'gm' (Geman-McClure) and 'cauchy' are provided.
    cutoff : float, optional
        Cut-off value for defining outliers based on fitting residuals.
        Will be passed to the outlier_condition_func.
        Typical example: ``|residuals| > cut_off x standard_deviation``
    outlier_condition_func : callable, optional
        A function with args and returns as follows::

            is_an_outlier = outlier_condition_func(residuals, log_residuals,
                pred_sig, design_matrix, leverages, C, cutoff)

    Notes
    -----
    Robust weights are calculated specifically for the WLS problem, i.e. the
    usual form of the WLS problem is accounted for when defining these new
    weights, see :footcite:p:`Collier2015`. On the second-to-last iteration,
    OLS is performed without outliers. On the last iteration, WLS is performed
    without outliers.

    References
    ----------
    .. footbibliography::

    """
    # check if M-estimator is valid (defined in this function)
    if m_est not in ["gm", "cauchy"]:
        raise ValueError("unknown M-estimator choice")

    # min 4 iters: (1) WLS (2) WLS with M-weights (3) clean OLS (3) clean WLS
    if total_idx < 4:
        raise ValueError("require 4+ iterations")

    p, N = design_matrix.shape[-1], data.shape[-1]
    if N <= p:
        raise ValueError("Fewer data points than parameters.")
    factor = 1.4826 * np.sqrt(N / (N - p))

    # handle potential zeros
    pred_sig[pred_sig < MIN_POSITIVE_SIGNAL] = MIN_POSITIVE_SIGNAL

    # calculate quantities needed for C and w
    log_pred_sig = np.log(pred_sig)
    residuals = data - pred_sig
    log_data = np.log(data)
    log_residuals = log_data - log_pred_sig
    z = pred_sig * log_residuals

    # IRLLS paper eq9 corrected (weights for log_residuals^2 are pred_sig^2)
    C = (
        factor
        * np.median(np.abs(z - np.median(z, axis=-1)[..., None]), axis=-1)[..., None]
    )
    C[C == 0] = np.nanmedian(C)  # C could be 0, if all signals = min_signal

    # NOTE: if more M-estimators are added, please update the docs!
    if m_est == "gm":
        w = (C / pred_sig) ** 2 / ((C / pred_sig) ** 2 + log_residuals**2) ** 2
    if m_est == "cauchy":
        w = C**2 / ((C / pred_sig) ** 2 + log_residuals**2)

    robust = None

    if idx == total_idx - 1:  # OLS without outliers
        cond = outlier_condition_func(
            residuals, log_residuals, pred_sig, design_matrix, leverages, C, cutoff
        )
        robust = np.logical_not(cond)

        w[~robust] = 0.0
        w[robust] = 1.0

    if idx == total_idx:  # WLS without outliers
        robust = last_robust
        w[~robust] = 0.0
        w[robust] = pred_sig[robust == 1] ** 2

    w[np.isinf(w)] = 0
    w[np.isnan(w)] = 0

    return w, robust


def weights_method_nlls_m_est(
    data,
    pred_sig,
    design_matrix,
    leverages,
    idx,
    total_idx,
    last_robust,
    *,
    m_est="gm",
    cutoff=3,
    outlier_condition_func=simple_cutoff,
):
    """Calculate M-estimator weights for NLLS model.

    Parameters
    ----------
    data : ndarray
        The measured signal.
    pred_sig : ndarray
        The predicted signal, given a previous fit.
        Has the same shape as data.
    design_matrix : ndarray (g, ...)
        Design matrix holding the covariants used to solve for the
        regression coefficients.
    leverages : ndarray
        The leverages (diagonal of the 'hat matrix') of the fit.
    idx : int
        The current iteration number.
    total_idx : int
        The total number of iterations.
    last_robust : ndarray
        True for inlier indices and False for outlier indices. Must have the
        same shape as data.
    m_est : str, optional.
        M-estimator weighting scheme to use. Currently,
        'gm' (Geman-McClure) and 'cauchy' are provided.
    cutoff : float, optional
        Cut-off value for defining outliers based on fitting residuals.
        Will be passed to the outlier_condition_func.
        Typical example: ``|residuals| > cut_off x standard_deviation``
    outlier_condition_func : callable, optional
        A function with args and returns as follows::

            is_an_outlier = outlier_condition_func(residuals, log_residuals,
                pred_sig, design_matrix, leverages, C, cutoff)

    Notes
    -----
    Robust weights are calculated specifically for the NLLS problem.
    On the last iteration, NLLS is performed without outliers.

    """
    # check if M-estimator is valid (defined in this function)
    if m_est not in ["gm", "cauchy"]:
        raise ValueError("unknown M-estimator choice")

    # min 3 iters: (1) NLLS (2) NLLS with M-weights (3) clean NLLS
    if total_idx < 3:
        raise ValueError("require 3+ iterations")

    p, N = design_matrix.shape[-1], data.shape[-1]
    if N <= p:
        raise ValueError("Fewer data points than parameters.")
    factor = 1.4826 * np.sqrt(N / (N - p))

    # handle potential zeros
    pred_sig[pred_sig < MIN_POSITIVE_SIGNAL] = MIN_POSITIVE_SIGNAL

    # calculate quantities needed for C and w
    log_pred_sig = np.log(pred_sig)
    residuals = data - pred_sig
    log_data = np.log(data)
    log_residuals = log_data - log_pred_sig

    C = (
        factor
        * np.median(np.abs(residuals - np.median(residuals)[..., None]), axis=-1)[
            ..., None
        ]
    )
    C[C == 0] = np.nanmedian(C)  # C could be 0, if all signals = min_signal

    # NOTE: if more M-estimators are added, please update the docs!
    if m_est == "gm":
        w = C**2 / (C**2 + residuals**2) ** 2
    if m_est == "cauchy":
        w = C**2 / (C**2 + residuals**2)

    robust = None

    if idx == total_idx:
        cond = outlier_condition_func(
            residuals, log_residuals, pred_sig, design_matrix, leverages, C, cutoff
        )
        robust = np.logical_not(cond)

        w[~robust] = 0.0
        w[robust] = 1.0

    w[np.isinf(w)] = 0
    w[np.isnan(w)] = 0

    return w, robust



import importlib
import sys
import warnings

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal, assert_raises
import pytest

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.shm import tournier07_legacy_msg
from dipy.utils.optpkg import optional_package

tf, have_tf, _ = optional_package("tensorflow", min_version="2.18.0")
torch, have_torch, _ = optional_package("torch", min_version="2.2.0")
have_nn = have_tf or have_torch
BACKENDS = [
    backend
    for backend, available in [("tf", have_tf), ("torch", have_torch)]
    if available
]


@pytest.mark.skipif(not have_nn, reason="Requires TensorFlow or Torch")
def test_default_weights(monkeypatch):
    input_arr = np.expand_dims(
        np.array(
            [
                1.15428471,
                0.37460899,
                -0.16743798,
                -0.02638639,
                -0.02587842,
                -0.24743459,
                -0.11091634,
                -0.01974129,
                -0.03463564,
                0.04234652,
                0.00909119,
                -0.02181194,
                -0.01141419,
                0.06747056,
                -0.02881568,
                0.0037776,
                0.02069041,
                0.01655271,
                -0.00958642,
                0.0103591,
                -0.00579612,
                0.00559265,
                0.00311974,
                0.0067629,
                0.00140297,
                0.01844978,
                -0.00551951,
                0.02215372,
                0.00186543,
                -0.01057652,
                0.00189625,
                -0.01114438,
                0.00509697,
                -0.00150783,
                0.01585437,
                0.00256389,
                0.00196107,
                -0.0108544,
                0.01143742,
                -0.00547229,
                -0.01040528,
                0.0114365,
                -0.02261801,
                0.00452243,
                0.0015014,
            ]
        ),
        axis=0,
    )

    target_arr = np.expand_dims(
        np.array(
            [
                0.17804961,
                -0.18878266,
                0.02026339,
                -0.0100488,
                -0.04045521,
                0.20264171,
                -0.16151273,
                0.00508221,
                0.01158303,
                0.03848331,
                0.04242867,
                -0.00493216,
                -0.05138939,
                0.03944791,
                -0.06210141,
                -0.01777741,
                0.00032369,
                -0.00781484,
                0.02685455,
                0.00617174,
                -0.01357785,
                0.0112316,
                0.02457713,
                -0.00307974,
                -0.00110319,
                -0.0274653,
                0.01606723,
                -0.05088685,
                0.0017358,
                -0.00533427,
                -0.00785866,
                0.00529946,
                0.00624491,
                0.00682212,
                -0.00551173,
                -0.00760572,
                -0.00145562,
                0.02271283,
                0.0238023,
                -0.01574752,
                0.00853913,
                -0.00715324,
                0.02677651,
                0.01718479,
                -0.01433261,
            ]
        ),
        axis=0,
    )

    for backend in BACKENDS:
        monkeypatch.setenv("DIPY_NN_BACKEND", backend)
        with warnings.catch_warnings():
            msg = ".*uses TensorFlow.*install PyTorch.*"
            warnings.filterwarnings("ignore", message=msg, category=DeprecationWarning)
            dipy_nn = importlib.reload(sys.modules["dipy.nn"])
            resdnn = dipy_nn.histo_resdnn

        resdnn_model = resdnn.HistoResDNN()
        resdnn_model.fetch_default_weights()
        results_arr = resdnn_model._HistoResDNN__predict(input_arr)
        assert_almost_equal(results_arr, target_arr, decimal=6)


@pytest.mark.skipif(not have_nn, reason="Requires TensorFlow or Torch")
def test_predict_shape_and_masking(monkeypatch):
    dwi_fname, bval_fname, bvec_fname = get_fnames(name="stanford_hardi")
    data, _ = load_nifti(dwi_fname)
    data = np.squeeze(data)
    bvals, bvecs = read_bvals_bvecs(bval_fname, bvec_fname)
    gtab = gradient_table(bvals, bvecs=bvecs)

    mask = np.zeros(data.shape[0:3], dtype=bool)
    mask[38:40, 45:50, 35:40] = 1

    for backend in BACKENDS:
        monkeypatch.setenv("DIPY_NN_BACKEND", backend)
        with warnings.catch_warnings():
            msg = ".*uses TensorFlow.*install PyTorch.*"
            warnings.filterwarnings("ignore", message=msg, category=DeprecationWarning)
            dipy_nn = importlib.reload(sys.modules["dipy.nn"])
            resdnn = dipy_nn.histo_resdnn

        resdnn_model = resdnn.HistoResDNN()
        resdnn_model.fetch_default_weights()
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=tournier07_legacy_msg,
                category=PendingDeprecationWarning,
            )
            results_arr = resdnn_model.predict(data, gtab, mask=mask)
        results_pos = np.sum(results_arr, axis=-1, dtype=bool)
        assert_equal(mask, results_pos)
        assert_equal(results_arr.shape[-1], 45)


@pytest.mark.skipif(not have_nn, reason="Requires TensorFlow or Torch")
def test_wrong_sh_order_weights(monkeypatch):
    for backend in BACKENDS:
        monkeypatch.setenv("DIPY_NN_BACKEND", backend)
        with warnings.catch_warnings():
            msg = ".*uses TensorFlow.*install PyTorch.*"
            warnings.filterwarnings("ignore", message=msg, category=DeprecationWarning)
            dipy_nn = importlib.reload(sys.modules["dipy.nn"])
            resdnn = dipy_nn.histo_resdnn

        resdnn_model = resdnn.HistoResDNN(sh_order_max=6)
        fetch_model_weights_path = get_fnames(name=f"histo_resdnn_{backend}_weights")
        if backend == "torch":
            assert_raises(
                RuntimeError, resdnn_model.load_model_weights, fetch_model_weights_path
            )
        else:  # tf
            assert_raises(
                ValueError, resdnn_model.load_model_weights, fetch_model_weights_path
            )


@pytest.mark.skipif(not have_nn, reason="Requires TensorFlow or Torch")
def test_wrong_sh_order_input(monkeypatch):
    for backend in BACKENDS:
        monkeypatch.setenv("DIPY_NN_BACKEND", backend)
        with warnings.catch_warnings():
            msg = ".*uses TensorFlow.*install PyTorch.*"
            warnings.filterwarnings("ignore", message=msg, category=DeprecationWarning)
            dipy_nn = importlib.reload(sys.modules["dipy.nn"])
            resdnn = dipy_nn.histo_resdnn

        resdnn_model = resdnn.HistoResDNN()
        fetch_model_weights_path = get_fnames(name=f"histo_resdnn_{backend}_weights")
        resdnn_model.load_model_weights(fetch_model_weights_path)
        assert_raises(ValueError, resdnn_model._HistoResDNN__predict, np.zeros((1, 28)))



import warnings

import numpy as np
import numpy.testing as npt

from dipy.core.sphere import HemiSphere, unit_octahedron
from dipy.data import default_sphere, get_sphere
from dipy.direction.pmf import SHCoeffPmfGen, SimplePmfGen
from dipy.reconst import shm
from dipy.testing.decorators import set_random_number_generator

response = (np.array([1.5e3, 0.3e3, 0.3e3]), 1)


@set_random_number_generator()
def test_pmf_val(rng):
    sphere = get_sphere(name="symmetric724")
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        pmfgen = SHCoeffPmfGen(rng.random([2, 2, 2, 28]), sphere, None)
    point = np.array([1, 1, 1], dtype="float")

    out = np.ones(len(sphere.vertices))
    for idx in [0, 5, 15, -1]:
        pmf = pmfgen.get_pmf(point)
        pmf_2 = pmfgen.get_pmf(point, out)

        npt.assert_array_almost_equal(pmf, out)
        npt.assert_array_almost_equal(pmf, pmf_2)
        # Create a direction vector close to the vertex idx
        xyz = sphere.vertices[idx] + rng.random([3]) / 100
        pmf_idx = pmfgen.get_pmf_value(point, xyz)
        # Test that the pmf sampled for the direction xyz is correct
        npt.assert_array_almost_equal(pmf[idx], pmf_idx)


def test_pmf_from_sh():
    sphere = HemiSphere.from_sphere(unit_octahedron)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=shm.descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        pmfgen = SHCoeffPmfGen(np.ones([2, 2, 2, 28]), sphere, None)

    out = np.zeros(len(sphere.vertices))
    # Test that the pmf is greater than 0 for a valid point
    pmf = pmfgen.get_pmf(np.array([0, 0, 0], dtype="float"))
    out = pmfgen.get_pmf(np.array([0, 0, 0], dtype="float"), out)
    npt.assert_equal(np.sum(pmf) > 0, True)
    npt.assert_array_almost_equal(pmf, out)

    # Test that the pmf is 0 for invalid Points
    npt.assert_array_equal(
        pmfgen.get_pmf(np.array([-1, 0, 0], dtype="float")),
        np.zeros(len(sphere.vertices)),
    )
    npt.assert_array_equal(
        pmfgen.get_pmf(np.array([0, 0, 10], dtype="float")),
        np.zeros(len(sphere.vertices)),
    )


def test_pmf_from_array():
    sphere = HemiSphere.from_sphere(unit_octahedron)
    pmfgen = SimplePmfGen(np.ones([2, 2, 2, len(sphere.vertices)]), sphere)

    out = np.zeros(len(sphere.vertices))
    # Test that the pmf is greater than 0 for a valid point
    pmf = pmfgen.get_pmf(np.array([0, 0, 0], dtype="float"))
    out = pmfgen.get_pmf(np.array([0, 0, 0], dtype="float"), out)
    npt.assert_equal(np.sum(pmf) > 0, True)
    npt.assert_array_almost_equal(pmf, out)

    # Test that the pmf is 0 for invalid Points
    npt.assert_array_equal(
        pmfgen.get_pmf(np.array([-1, 0, 0], dtype=float)),
        np.zeros(len(sphere.vertices)),
    )
    npt.assert_array_equal(
        pmfgen.get_pmf(np.array([0, 0, 10], dtype=float)),
        np.zeros(len(sphere.vertices)),
    )

    # Test ValueError for non matching pmf and sphere
    npt.assert_raises(
        ValueError,
        lambda: SimplePmfGen(np.ones([2, 2, 2, len(sphere.vertices)]), default_sphere),
    )



"""
==============================================================================
Reconstruction of the diffusion signal with the correlation tensor model (CTI)
==============================================================================

Correlation Tensor MRI (CTI) is a method that uses double diffusion
encoding data to resolve sources of kurtosis. It is similar to the Q-space
Trajectory Imaging method
(see :ref:`sphx_glr_examples_built_reconstruction_reconst_qti.py`)
:footcite:p:`NetoHenriques2020`. However, in addition to the kurtosis sources
associated with diffusion variance across compartments (``K_aniso`` and
``K_iso``, which are related to microscopic anisotropy and the variance of the
mean diffusivities of compartments, respectively), CTI also measures K_micro.
This quantifies non-Gaussian diffusion effects that deviate from the multiple
Gaussian component tissue representation, such as restricted diffusion,
exchange, and structural disorder in compartments like cross-sectional
variance :footcite:p:`Novello2022`, :footcite:p:`Alves2022`.

Although the CorrelationTensorModel and the DiffusionKurtosisTensorFit may
share some similarities, they have significantly different representations for
the diffusion-weighted signal. This difference leads to the necessity for a
unique ``design matrix`` specifically for CTI.
The CorrelationTensorModel expresses the diffusion-weighted signal as:

.. math::

    \\begin{align}
    \\log E_{\\Delta}(q_1, q_2) &= \\left(q_{1i}q_{1j}
    + q_{2i}q_{2j}\\right) \\Delta D_{ij} \\
    &+ q_{1i}q_{2j}Q_{ij} \\
    &+ \\frac{1}{6} \\left( q_{1i}q_{1j}q_{1k}q_{1l}
    + q_{2i}q_{2j}q_{2k}q_{2l} \\right) \\
    &\\quad \\times \\Delta^2 D^2 W_{ijkl} \\
    &+ \\frac{1}{4} q_{1i}q_{1j}q_{2k}q_{2l}Z_{ijkl} \\
    &+ \\frac{1}{6} \\left( q_{1i}q_{1j}q_{1k}q_{2l}
    + q_{2i}q_{2j}q_{2k}q_{1l} \\right) S_{ijkl} \\
    &+ O(q^6)
    \\end{align}

where: $\\Delta D_{ij}$ refers to the elements of the total diffusion
tensor (DT) and $W_{ijkl}$ are the elements of the total kurtosis tensor
(KT) and $D$ is the mean diffusivity and $Q_{ij}$ are the elements of a 2nd
order correlation tensor Q.
$Z_{ijkl}$ and $S_{ijkl}$ are the elements of the 4th order displacement
correlation tensors Z and S, respectively.

However it differentiates from the DiffusionKurtosis Model by calculating the
different sources of kurtosis.

In the following example we show how to fit the correlation tensor model on a
real life dataset and how to estimate correlation tensor based statistics.

First, we'll import all relevant modules.
"""

import matplotlib.pyplot as plt

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io import read_bvals_bvecs
from dipy.io.image import load_nifti
import dipy.reconst.cti as cti

###############################################################################
# For CTI analysis, data must be acquired using double diffusion encoding,
# taking into account different pairs of b-values and gradient directions
# between the two diffusion epochs, as discussed by Henriques et al. in Magn
# Reson Med 2021 [NetoHe2021]._
# To run CTI we need to have separate b-value and b-vector files for each DDE
# diffusion epoch. In this tutorial, a sample DDE dataset and respective
# b-value and b-vector files for the two epochs are fetched. If you want to
# process your own DDE data compatible with CTI, you will have to change the
# below lines of code to add the paths of your data.
# The users should also ensure that the data is formatted correctly for the
# CTI analysis they are performing.

fdata, fbvals1, fbvecs1, fbvals2, fbvecs2, fmask = get_fnames(name="cti_rat1")
data, affine = load_nifti(fdata)
bvals1, bvecs1 = read_bvals_bvecs(fbvals1, fbvecs1)
bvals2, bvecs2 = read_bvals_bvecs(fbvals2, fbvecs2)

###############################################################################
# In this example, the function load_nifti is used to load the CTI data saved
# in filefdata and returns the data as a nibabel Nifti1Image object along with
# the affine transformation. The b-values and b-vectors for two different
# gradient tables are loaded from ``bvals1.bval`` and ``bvec1.bvec``, and
# ``bvals2.bval`` and ``bvec2.bvec``, respectively, using the
# ``read_bvals_bvecs`` function. For CTI reconstruction in DIPY, we need to
# define the b-values and b-vectors for each diffusion epoch in separate
# gradient tables, as done in the above line of code.

gtab1 = gradient_table(bvals1, bvecs=bvecs1)
gtab2 = gradient_table(bvals2, bvecs=bvecs2)

###############################################################################
# Before fitting the data, we perform some data pre-processing. We first
# compute a brain mask to avoid unnecessary calculations on the background
# of the image.

mask, mask_affine = load_nifti(fmask)

###############################################################################
# Now that we've loaded the data and generated the two gradient tables we can
# go forward with CTI fitting. For this, the CTI model is first defined for
# GradientTable objects gtab1 and gtab2 by instantiating the
# CorrelationTensorModel object in the following way:

ctimodel = cti.CorrelationTensorModel(gtab1, gtab2)

###############################################################################
# To fit the data using the defined model object, we call the fit function of
# this object.

ctifit = ctimodel.fit(data, mask=mask)

###############################################################################
# The fit method for the CTI model produces a CorrelationTensorFit object,
# which contains the attributes of both the DKI and DTI models. Given that CTI
# is a built upon DKI, which itself extends the DTI model, the
# CorrelationTensorFit instance captures a comprehensive set of parameters and
# attributes from these underlying models.
#
# For instance, the CTI model inherently estimates all DTI and DKI statistics,
# such as mean, axial, and radial diffusivities (MD, AD, RD) as well as the
# mean, axial, and radial kurtosis (MK, AK, RK).
# To better illustrate the extraction of main DTI/DKI parameters using the CTI
# model, consider the following lines of code:

AD = ctifit.ad
MD = ctifit.md
RD = ctifit.rd
MK = ctifit.mk()
AK = ctifit.ak()
RK = ctifit.rk()

###############################################################################
# However, in addition to these metrics, CTI also provides unique sources of
# information, not available in DTI and DKI. Below we draw a feature map of the
# 3 different sources of kurtosis which can exclusively be calculated from the
# CTI model.

kiso_map = ctifit.K_iso
kaniso_map = ctifit.K_aniso
kmicro_map = ctifit.K_micro

slice_idx = 0
fig, axarr = plt.subplots(1, 3, figsize=(15, 5))

axarr[0].imshow(kiso_map[:, :, slice_idx], cmap="gray", origin="lower", vmin=0, vmax=1)
axarr[0].set_title("Kiso Map")

axarr[1].imshow(
    kaniso_map[:, :, slice_idx], cmap="gray", origin="lower", vmin=0, vmax=1
)
axarr[1].set_title("Kaniso Map")

axarr[2].imshow(
    kmicro_map[:, :, slice_idx], cmap="gray", origin="lower", vmin=0, vmax=1
)
axarr[2].set_title("Kmicro Map")

plt.tight_layout()
plt.show()

###############################################################################
# References
# ----------
#
# .. footbibliography::
#



"""
==========================================
Symmetric Diffeomorphic Registration in 2D
==========================================
This example explains how to register 2D images using the Symmetric
Normalization (SyN) algorithm proposed by :footcite:t:`Avants2008` (also
implemented in the ANTs software :footcite:p:`Avants2009`)

We will perform the classic Circle-To-C experiment for diffeomorphic
registration
"""

import numpy as np

import dipy.align.imwarp as imwarp
from dipy.align.imwarp import SymmetricDiffeomorphicRegistration
from dipy.align.metrics import CCMetric, SSDMetric
from dipy.data import get_fnames
from dipy.io.image import load_nifti_data
from dipy.segment.mask import median_otsu
from dipy.viz import regtools

fname_moving = get_fnames(name="reg_o")
fname_static = get_fnames(name="reg_c")

moving = np.load(fname_moving)
static = np.load(fname_static)

###############################################################################
# To visually check the overlap of the static image with the transformed moving
# image, we can plot them on top of each other with different channels to see
# where the differences are located

regtools.overlay_images(
    static,
    moving,
    title0="Static",
    title_mid="Overlay",
    title1="Moving",
    fname="input_images.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Input images.
#
#
#
# We want to find an invertible map that transforms the moving image (circle)
# into the static image (the C letter).
#
# The first decision we need to make is what similarity metric is appropriate
# for our problem. In this example we are using two binary images, so the Sum
# of Squared Differences (SSD) is a good choice.

dim = static.ndim
metric = SSDMetric(dim)

###############################################################################
# Now we define an instance of the registration class. The SyN algorithm uses
# a multi-resolution approach by building a Gaussian Pyramid. We instruct the
# registration instance to perform at most $[n_0, n_1, ..., n_k]$ iterations
# at each level of the pyramid. The 0-th level corresponds to the finest
# resolution.

level_iters = [200, 100, 50, 25]

sdr = SymmetricDiffeomorphicRegistration(metric, level_iters=level_iters, inv_iter=50)

###############################################################################
# Now we execute the optimization, which returns a DiffeomorphicMap object,
# that can be used to register images back and forth between the static and
# moving domains

mapping = sdr.optimize(static, moving)

###############################################################################
# It is a good idea to visualize the resulting deformation map to make sure
# the result is reasonable (at least, visually)

regtools.plot_2d_diffeomorphic_map(mapping, delta=10, fname="diffeomorphic_map.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Deformed lattice under the resulting diffeomorphic map.
#
#
#
# Now let's warp the moving image and see if it gets similar to the static
# image

warped_moving = mapping.transform(moving, interpolation="linear")
regtools.overlay_images(
    static,
    warped_moving,
    title0="Static",
    title_mid="Overlay",
    title1="Warped moving",
    fname="direct_warp_result.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Moving image transformed under the (direct) transformation in green on top
# of the static image (in red).
#
#
#
# And we can also apply the inverse mapping to verify that the warped static
# image is similar to the moving image

warped_static = mapping.transform_inverse(static, interpolation="linear")
regtools.overlay_images(
    warped_static,
    moving,
    title0="Warped static",
    title_mid="Overlay",
    title1="Moving",
    fname="inverse_warp_result.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Static image transformed under the (inverse) transformation in red on top
# of the moving image (in green).
#
#
#
# Now let's register a couple of slices from a b0 image using the Cross
# Correlation metric. Also, let's inspect the evolution of the registration.
# To do this we will define a function that will be called by the registration
# object at each stage of the optimization process. We will draw the current
# warped images after finishing each resolution.


def callback_CC(sdr, status):
    # Status indicates at which stage of the optimization we currently are
    # For now, we will only react at the end of each resolution of the scale
    # space
    if status == imwarp.RegistrationStages.SCALE_END:
        # get the current images from the metric
        wmoving = sdr.metric.moving_image
        wstatic = sdr.metric.static_image
        # draw the images on top of each other with different colors
        regtools.overlay_images(
            wmoving,
            wstatic,
            title0="Warped moving",
            title_mid="Overlay",
            title1="Warped static",
        )


###############################################################################
# Now we are ready to configure and run the registration. First load the data

t1_name, b0_name = get_fnames(name="syn_data")
data = load_nifti_data(b0_name)

###############################################################################
# We first remove the skull from the b0 volume

b0_mask, mask = median_otsu(data, median_radius=4, numpass=4)

###############################################################################
# And select two slices to try the 2D registration

static = b0_mask[:, :, 40]
moving = b0_mask[:, :, 38]

###############################################################################
# After loading the data, we instantiate the Cross-Correlation metric. The
# metric receives three parameters: the dimension of the input images, the
# standard deviation of the Gaussian Kernel to be used to regularize the
# gradient and the radius of the window to be used for evaluating the local
# normalized cross correlation.

sigma_diff = 3.0
radius = 4
metric = CCMetric(2, sigma_diff=sigma_diff, radius=radius)

###############################################################################
# Let's use a scale space of 3 levels

level_iters = [100, 50, 25]
sdr = SymmetricDiffeomorphicRegistration(metric, level_iters=level_iters)
sdr.callback = callback_CC

###############################################################################
# And execute the optimization

mapping = sdr.optimize(static, moving)

warped = mapping.transform(moving)

###############################################################################
# We can see the effect of the warping by switching between the images before
# and after registration

regtools.overlay_images(
    static,
    moving,
    title0="Static",
    title_mid="Overlay",
    title1="Moving",
    fname="t1_slices_input.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Input images.

regtools.overlay_images(
    static,
    warped,
    title0="Static",
    title_mid="Overlay",
    title1="Warped moving",
    fname="t1_slices_res.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Moving image transformed under the (direct) transformation in green on top
# of the static image (in red).
#
#
#
# And we can apply the inverse warping too

inv_warped = mapping.transform_inverse(static)
regtools.overlay_images(
    inv_warped,
    moving,
    title0="Warped static",
    title_mid="Overlay",
    title1="moving",
    fname="t1_slices_res2.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Static image transformed under the (inverse) transformation in red on top
# of the moving image (in green).
#
#
#
# Finally, let's see the deformation

regtools.plot_2d_diffeomorphic_map(mapping, delta=5, fname="diffeomorphic_map_b0s.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Deformed lattice under the resulting diffeomorphic map.
#
#
# References
# ----------
#
# .. footbibliography::
#



cimport numpy as cnp

from dipy.direction.pmf cimport PmfGen
from dipy.tracking.direction_getter cimport DirectionGetter


cdef int closest_peak(cnp.ndarray[cnp.float_t, ndim=2] peak_dirs,
                      double[::1] direction, double cos_similarity)


cdef class BasePmfDirectionGetter(DirectionGetter):

    cdef:
        dict _pf_kwargs
        double pmf_threshold
        double cos_similarity
        int len_pmf
        object sphere
        PmfGen pmf_gen

    cpdef cnp.ndarray[cnp.float_t, ndim=2] initial_direction(
        self,
        double[::1] point)

    cdef double* _get_pmf(self, double[::1] point) nogil
    cdef int get_direction_c(self, double[::1] point, double[::1] direction)


cdef class BaseDirectionGetter(BasePmfDirectionGetter):

    pass


cdef class PmfGenDirectionGetter(BasePmfDirectionGetter):

    pass


cdef class ClosestPeakDirectionGetter(PmfGenDirectionGetter):

    pass



"""
=====================
Simple volume slicing
=====================

Here we present an example for visualizing slices from 3D images.

Let's start by importing the relevant modules.
"""

import os

from dipy.data import fetch_bundles_2_subjects
from dipy.io.image import load_nifti, load_nifti_data
from dipy.viz import actor, ui, window

###############################################################################
# Let's download and load a T1.

fetch_bundles_2_subjects()

fname_t1 = os.path.join(
    os.path.expanduser("~"),
    ".dipy",
    "exp_bundles_and_maps",
    "bundles_2_subjects",
    "subj_1",
    "t1_warped.nii.gz",
)

data, affine = load_nifti(fname_t1)

###############################################################################
# Create a Scene object which holds all the actors which we want to visualize.

scene = window.Scene()
scene.background((0.5, 0.5, 0.5))

###############################################################################
# Render slices from T1 with a specific value range
# =================================================
#
# The T1 has usually a higher range of values than what can be visualized in an
# image. We can set the range that we would like to see.

mean, std = data[data > 0].mean(), data[data > 0].std()
value_range = (mean - 0.5 * std, mean + 1.5 * std)

###############################################################################
# The ``slice`` function will read data and resample the data using an affine
# transformation matrix. The default behavior of this function is to show the
# middle slice of the last dimension of the resampled data.

slice_actor = actor.slicer(data, affine=affine, value_range=value_range)

###############################################################################
# The ``slice_actor`` contains an axial slice.

scene.add(slice_actor)

###############################################################################
# The same actor can show any different slice from the given data using its
# ``display`` function. However, if we want to show multiple slices we need to
# copy the actor first.

slice_actor2 = slice_actor.copy()

###############################################################################
# Now we have a new ``slice_actor`` which displays the middle slice of the
# sagittal plane.

slice_actor2.display(x=slice_actor2.shape[0] // 2, y=None, z=None)

scene.add(slice_actor2)

scene.reset_camera()
scene.zoom(1.4)

###############################################################################
# In order to interact with the data you will need to uncomment the line below.

# window.show(scene, size=(600, 600), reset_camera=False)

###############################################################################
# Otherwise, you can save a screenshot using the following command.

window.record(scene=scene, out_path="slices.png", size=(600, 600), reset_camera=False)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Simple slice viewer.
#
#
# Render slices from FA with your colormap
# ========================================
#
# It is also possible to set the colormap of your preference. Here we are
# loading an FA image and showing it in a non-standard way using an HSV
# colormap.

fname_fa = os.path.join(
    os.path.expanduser("~"),
    ".dipy",
    "exp_bundles_and_maps",
    "bundles_2_subjects",
    "subj_1",
    "fa_1x1x1.nii.gz",
)

fa = load_nifti_data(fname_fa)

lut = actor.colormap_lookup_table(
    scale_range=(fa.min(), fa.max() * 0.8),
    hue_range=(0.4, 1.0),
    saturation_range=(1, 1.0),
    value_range=(0.0, 1.0),
)

###############################################################################
# This is because the lookup table is applied in the slice after interpolating
# to (0, 255).

fa_actor = actor.slicer(fa, affine=affine, lookup_colormap=lut)

scene.clear()
scene.add(fa_actor)

scene.reset_camera()
scene.zoom(1.4)

# window.show(scene, size=(600, 600), reset_camera=False)

window.record(
    scene=scene, out_path="slices_lut.png", size=(600, 600), reset_camera=False
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Simple slice viewer with an HSV colormap.
#
#
#
# Now we would like to add the ability to click on a voxel and show its value
# on a panel in the window. The panel is a UI element which requires access to
# different areas of the visualization pipeline and therefore we don't
# recommend using it with ``window.show``. The more appropriate way is to use
# the ``ShowManager`` object, which allows accessing the pipeline in different
# areas.

show_m = window.ShowManager(scene=scene, size=(1200, 900))
show_m.initialize()

###############################################################################
# We'll start by creating the panel and adding it to the ``ShowManager``

label_position = ui.TextBlock2D(text="Position:")
label_value = ui.TextBlock2D(text="Value:")

result_position = ui.TextBlock2D(text="")
result_value = ui.TextBlock2D(text="")

panel_picking = ui.Panel2D(
    size=(250, 125), position=(20, 20), color=(0, 0, 0), opacity=0.75, align="left"
)

panel_picking.add_element(label_position, (0.1, 0.55))
panel_picking.add_element(label_value, (0.1, 0.25))

panel_picking.add_element(result_position, (0.45, 0.55))
panel_picking.add_element(result_value, (0.45, 0.25))

scene.add(panel_picking)

###############################################################################
# Add a left-click callback to the slicer. Also disable interpolation so you
# can see what you are picking.


def left_click_callback(obj, ev):
    """Get the value of the clicked voxel and show it in the panel."""
    event_pos = show_m.iren.GetEventPosition()

    obj.picker.Pick(event_pos[0], event_pos[1], 0, scene)

    i, j, k = obj.picker.GetPointIJK()
    result_position.message = f"({str(i)}, {str(j)}, {str(k)})"
    result_value.message = f"{data[i, j, k]:.8f}"


fa_actor.SetInterpolate(False)
fa_actor.AddObserver("LeftButtonPressEvent", left_click_callback, 1.0)

# show_m.start()

###############################################################################
# Create a mosaic
# ===============
#
# By using the ``copy`` and ``display`` method of the ``slice_actor`` it
# becomes easy and efficient to create a mosaic of all the slices.
#
# So, let's clear the scene and change the projection from perspective to
# parallel. We'll also need a new show manager and an associated callback.

scene.clear()
scene.projection("parallel")

result_position.message = ""
result_value.message = ""

show_m_mosaic = window.ShowManager(scene=scene, size=(1200, 900))
show_m_mosaic.initialize()


def left_click_callback_mosaic(obj, ev):
    """Get the value of the clicked voxel and show it in the panel."""
    event_pos = show_m_mosaic.iren.GetEventPosition()

    obj.picker.Pick(event_pos[0], event_pos[1], 0, scene)

    i, j, k = obj.picker.GetPointIJK()
    result_position.message = f"({str(i)}, {str(j)}, {str(k)})"
    result_value.message = f"{data[i, j, k]:.8f}"


###############################################################################
# Now we need to create two nested for loops which will set the positions of
# the grid of the mosaic and add the new actors to the scene. We are going
# to use 15 columns and 10 rows but you can adjust those with your datasets.

cnt = 0

X, Y, Z = slice_actor.shape[:3]

rows = 10
cols = 15
border = 10

for j in range(rows):
    for i in range(cols):
        slice_mosaic = slice_actor.copy()
        slice_mosaic.display(z=cnt)
        slice_mosaic.SetPosition(
            (X + border) * i, 0.5 * cols * (Y + border) - (Y + border) * j, 0
        )
        slice_mosaic.SetInterpolate(False)
        slice_mosaic.AddObserver(
            "LeftButtonPressEvent", left_click_callback_mosaic, 1.0
        )
        scene.add(slice_mosaic)
        cnt += 1
        if cnt > Z:
            break
    if cnt > Z:
        break

scene.reset_camera()
scene.zoom(1.0)

# show_m_mosaic.ren.add(panel_picking)
# show_m_mosaic.start()

###############################################################################
# If you uncomment the two lines above, you will be able to move
# the mosaic up/down and left/right using the middle mouse button drag,
# zoom in/out using the scroll wheel, and pick voxels with left click.

window.record(scene=scene, out_path="mosaic.png", size=(900, 600), reset_camera=False)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# A mosaic of all the slices in the T1 volume.



"""
plotting functions
"""

from warnings import warn

import numpy as np

from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

plt, have_plt, _ = optional_package("matplotlib.pyplot")


@warning_for_keywords()
def compare_maps(
    fits,
    maps,
    *,
    transpose=None,
    fit_labels=None,
    map_labels=None,
    fit_kwargs=None,
    map_kwargs=None,
    filename=None,
):
    """Compare one or more scalar maps for different fits or models.

    Parameters
    ----------
    fits : list
        List of fits to be compared.
    maps : list
        Names of attributes to be compared.
        Default: 'rtop'.
    transpose : bool, optional
        If False, different fits are placed on different rows and different
        maps on different columns. If True, the order is transposed. If None,
        the figures are placed such that there are more columns than rows.
        Default: None.
    fit_labels : list, optional
        Labels for the different fitting routines. If None the fits are labeled
        by number.
        Default: None.
    map_labels : list, optional
        Labels for the different attributes. If None the attribute names are
        used.
        Default: None.
    fit_kwargs : list or dict, optional
        A dict or list of dicts with imshow options for each fitting routine.
        The dicts are passed to imshow as keyword-argument pairs.
        Default: {}.
    map_kwargs : list or dict, optional
        A dict or list of dicts with imshow options for each MAP-MRI scalar.
        The dicts are passed to imshow as keyword-argument pairs.
        Default: {}.
    filename : string, optional
        Filename where the image will be saved.
        Default: None.
    """
    fit_kwargs = fit_kwargs or {}
    map_kwargs = map_kwargs or {}

    if not have_plt:
        raise ValueError("matplotlib package needed for visualization.")

    fontsize = "large"
    xscale, yscale = 12, 10

    m = len(fits)
    n = len(maps)

    if transpose is None:
        transpose = m > n

    if fit_labels is None:
        fit_labels = [f"Fit {i + 1}" for i in range(m)]
    if map_labels is None:
        map_labels = maps

    if isinstance(fit_kwargs, dict):
        fit_kwargs = [fit_kwargs] * m
    if isinstance(map_kwargs, dict):
        map_kwargs = [map_kwargs] * n

    if transpose:
        fig, ax = plt.subplots(n, m, figsize=(xscale, yscale / m * n), squeeze=False)
        ax = ax.T
        for i in range(m):
            ax[i, 0].set_title(fit_labels[i], fontsize=fontsize)
        for j in range(n):
            ax[0, j].set_ylabel(map_labels[j], fontsize=fontsize)
    else:
        fig, ax = plt.subplots(m, n, figsize=(xscale, yscale / n * m), squeeze=False)
        for i in range(m):
            ax[i, 0].set_ylabel(fit_labels[i], fontsize=fontsize)
        for j in range(n):
            ax[0, j].set_title(map_labels[j], fontsize=fontsize)

    for i in range(m):
        for j in range(n):
            try:
                attr = getattr(fits[i], maps[j])
                if callable(attr):
                    attr = attr()
            except AttributeError:
                warn(f"Could not recover attribute {maps[j]}.", stacklevel=2)
                attr = np.zeros((2, 2))
            data = np.squeeze(np.array(attr, dtype=float)).T
            ax[i, j].imshow(
                data,
                interpolation="nearest",
                origin="lower",
                cmap="gray",
                **fit_kwargs[i],
                **map_kwargs[j],
            )
            ax[i, j].set_xticks([])
            ax[i, j].set_yticks([])
            ax[i, j].spines["top"].set_visible(False)
            ax[i, j].spines["right"].set_visible(False)
            ax[i, j].spines["bottom"].set_visible(False)
            ax[i, j].spines["left"].set_visible(False)

    fig.tight_layout()

    if filename:
        plt.savefig(filename)
    else:
        plt.show()


@warning_for_keywords()
def compare_qti_maps(
    gt,
    fit1,
    fit2,
    mask,
    *,
    maps=("fa", "ufa"),
    fitname=("QTI", "QTI+"),
    xlimits=([0, 1], [0.4, 1.5]),
    disprange=([0, 1], [0, 1]),
    slice=13,
):
    """Compare one or more qti derived maps obtained with
    different fitting routines.

    Parameters
    ----------
    gt : qti fit object
        The qti fit to be considered as ground truth
    fit1 : qti fit object
        First qti fit to be compared
    fit2 : qti fit object
        Second qti fit to be compared
    mask : np.ndarray
        Boolean array indicating which voxels to retain for comparing
        the values
    maps : array-like, optional
        QTI invariants to be compared
    fitname : array-like, optional
        Names of the used QTI fitting routines
    xlimits : array-like, optional
        X-Axis limits for the histograms visualization
    disprange : array-like, optional
        Display range for maps
    slice : int, optional
        Axial brain slice to be visualized
    """
    if not have_plt:
        raise ValueError("matplotlib package needed for visualization")

    n = len(maps)
    fig, ax = plt.subplots(n, 4, figsize=(12, 9))

    background = np.zeros(gt.S0_hat.shape[0:2])
    for i in range(n):
        for j in range(3):
            ax[i, j].imshow(background, cmap="gray")
            ax[i, j].set_xticks([])
            ax[i, j].set_yticks([])

    for k in range(n):
        ax[k, 0].imshow(
            np.rot90(getattr(gt, maps[k])[:, :, slice]),
            cmap="gray",
            vmin=disprange[k][0],
            vmax=disprange[k][1],
        )
        ax[k, 0].set_title("GROUND TRUTH")
        ax[k, 0].set_ylabel(maps[k], fontsize=20)

        ax[k, 1].imshow(
            np.rot90(getattr(fit1, maps[k])[:, :, slice]),
            cmap="gray",
            vmin=disprange[k][0],
            vmax=disprange[k][1],
        )
        ax[k, 1].set_title(fitname[0])

        ax[k, 2].imshow(
            np.rot90(getattr(fit2, maps[k])[:, :, slice]),
            cmap="gray",
            vmin=disprange[k][0],
            vmax=disprange[k][1],
        )
        ax[k, 2].set_title(fitname[1])

        ax[k, 3].hist(
            (getattr(fit1, maps[k])[mask, slice]).flatten(),
            density=True,
            bins=40,
            label=fitname[0],
        )
        ax[k, 3].hist(
            (getattr(fit2, maps[k])[mask, slice]).flatten(),
            density=True,
            bins=40,
            label=fitname[1],
            alpha=0.7,
        )
        ax[k, 3].hist(
            (getattr(gt, maps[k])[mask, slice]).flatten(),
            histtype="stepfilled",
            density=True,
            bins=40,
            label="GT",
            ec="k",
            alpha=1,
            linewidth=1.5,
            fc="None",
        )
        ax[k, 3].legend()
        ax[k, 3].set_title("VALUE DISTRIBUTION")
        ax[k, 3].set_xlim(xlimits[k])

    fig.tight_layout()
    plt.show()


def bundle_shape_profile(x, shape_profile, std):
    """Plot bundlewarp bundle shape profile.

    Parameters
    ----------
    x : np.ndarray
        Integer array containing x-axis
    shape_profile : np.ndarray
        Float array containing bundlewarp displacement magnitudes along the
        length of the bundle
    std : np.ndarray
        Float array containing standard deviations
    """
    fig, ax = plt.subplots(figsize=(8, 6), dpi=300)
    std_1 = shape_profile + std
    std_2 = shape_profile - std
    ax.plot(
        x, shape_profile, "-", label="Mean", color="Purple", linewidth=3, markersize=12
    )
    ax.fill_between(x, std_1, std_2, alpha=0.2, label="Std", color="Purple")

    plt.xticks(x)
    plt.ylim(0, max(std_1) + 2)

    plt.ylabel("Average Displacement")
    plt.xlabel("Segment Number")
    plt.title("Bundle Shape Profile")
    plt.legend(loc=2)
    plt.show()


def image_mosaic(
    images, *, ax_labels=None, ax_kwargs=None, figsize=None, filename=None
):
    """
    Draw a mosaic of 2D images using pyplot.imshow(). A colorbar is drawn
    beside each image.

    Parameters
    ----------
    images: list of ndarray
        Images to render.
    ax_labels: list of str, optional
        Label for each image.
    ax_kwargs: list of dictionaries, optional
        keyword arguments passed to imshow for each image. One dictionary per
        image.
    figsize: tuple of ints, optional
        Figure size.
    filename: str, optional
        When given, figure is saved to disk under this name.

    Returns
    -------
    fig: pyplot.Figure
        The figure.
    ax: pyplot.Axes or array of Axes
        The subplots for each image.
    """
    fig, ax = plt.subplots(1, len(images), figsize=figsize)

    aximages = []
    for it, (im, axe, kw) in enumerate(zip(images, ax, ax_kwargs)):
        aximages.append(axe.imshow(im, **kw))
        if ax_labels is not None:
            axe.set_title(ax_labels[it])

    for it, aximage in enumerate(aximages):
        fig.colorbar(aximage, ax=ax[it])

    if filename is not None:
        plt.savefig(filename)
    else:
        plt.show()

    return fig, ax
